# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

import math
import os
from pathlib import Path

import lightning.pytorch as pl
import numpy as np
import pyro
import pytest
import torch

from cellarium.ml import CellariumModule
from cellarium.ml.callbacks import VarianceMonitor
from cellarium.ml.models import ProbabilisticPCA, ProbabilisticPCAFromCLI
from cellarium.ml.transforms import NormalizeTotal
from cellarium.ml.utilities.data import collate_fn
from tests.common import BoringDataset


@pytest.fixture
def x_ng():
    n, g, k = 1000, 10, 3
    rng = np.random.default_rng(0)
    z_nk = rng.standard_normal(size=(n, k), dtype=np.float32)
    w_kg = rng.standard_normal(size=(k, g), dtype=np.float32)
    sigma = 0.6
    noise = sigma * rng.standard_normal(size=(n, g), dtype=np.float32)
    x_ng = z_nk @ w_kg + noise
    return x_ng


@pytest.mark.parametrize("ppca_flavor", ["marginalized", "linear_vae"])
@pytest.mark.parametrize("learn_mean", [False, True])
@pytest.mark.parametrize("minibatch", [False, True], ids=["fullbatch", "minibatch"])
def test_probabilistic_pca_multi_device(x_ng: np.ndarray, minibatch: bool, ppca_flavor: str, learn_mean: bool):
    n, g = x_ng.shape
    k = 3
    devices = int(os.environ.get("TEST_DEVICES", "1"))
    if learn_mean:
        x_mean_g = None
    else:
        x_mean_g = torch.as_tensor(x_ng.mean(axis=0))

    # dataloader
    batch_size = n // 2 if minibatch else n
    train_loader = torch.utils.data.DataLoader(
        BoringDataset(x_ng),
        batch_size=batch_size,
        shuffle=False,
        collate_fn=collate_fn,
    )
    # model
    pyro.clear_param_store()
    total_var = np.var(x_ng, axis=0).sum()
    w = np.sqrt(0.5 * total_var / (g * k))
    s = np.sqrt(0.5 * total_var / g)
    ppca = ProbabilisticPCA(
        n_cells=n,
        g_genes=g,
        k_components=k,
        ppca_flavor=ppca_flavor,
        mean_g=x_mean_g,
        W_init_scale=w,
        sigma_init_scale=s,
    )
    module = CellariumModule(
        ppca,
        optim_fn=torch.optim.Adam,
        optim_kwargs={"lr": 3e-2},
        scheduler_fn=torch.optim.lr_scheduler.CosineAnnealingLR,
        scheduler_kwargs={"T_max": 1000},  # one cycle
    )
    # trainer
    trainer = pl.Trainer(
        barebones=True,
        accelerator="cpu",
        devices=devices,
        max_steps=1000,
    )
    # fit
    trainer.fit(module, train_dataloaders=train_loader)

    # pca fit
    x_ng_centered = x_ng - x_ng.mean(axis=0)
    x_cov_gg = x_ng_centered.T @ x_ng_centered / n
    L_g, U_gg = np.linalg.eig(x_cov_gg)

    # total variance
    expected_total_var = np.var(x_ng, axis=0).sum()
    actual_total_var = ppca.W_variance + ppca.sigma_variance
    np.testing.assert_allclose(expected_total_var, actual_total_var, rtol=1e-3)

    # variance explained be each PC
    expected_explained_var = L_g[:k]
    actual_explained_var = ppca.L_k
    np.testing.assert_allclose(expected_explained_var, actual_explained_var, rtol=1e-3)

    # absolute cosine similarity between expected and actual PCs
    abs_cos_sim = torch.abs(
        torch.nn.functional.cosine_similarity(
            ppca.U_gk,
            torch.as_tensor(U_gg[:, :k]),
            dim=0,
        )
    )
    np.testing.assert_allclose(np.ones(k), abs_cos_sim, rtol=1e-3)


def test_variance_monitor(x_ng: np.ndarray):
    n, g = x_ng.shape
    k = 3
    # dataloader
    train_loader = torch.utils.data.DataLoader(
        BoringDataset(x_ng),
        batch_size=n // 2,
        collate_fn=collate_fn,
    )
    # model
    ppca = ProbabilisticPCA(n, g, k, "marginalized")
    module = CellariumModule(ppca)
    # trainer
    var_monitor = VarianceMonitor(total_variance=np.var(x_ng, axis=0).sum())
    trainer = pl.Trainer(
        accelerator="cpu",
        devices=1,
        max_steps=2,
        callbacks=[var_monitor],
        log_every_n_steps=1,
    )
    # fit
    trainer.fit(module, train_dataloaders=train_loader)


def test_load_from_checkpoint_multi_device(tmp_path: Path):
    n, g = 3, 2
    devices = int(os.environ.get("TEST_DEVICES", "1"))
    # dataloader
    train_loader = torch.utils.data.DataLoader(
        BoringDataset(np.arange(n * g).reshape(n, g)),
        collate_fn=collate_fn,
    )
    # model
    init_args = {
        "n_cells": n,
        "g_genes": g,
        "k_components": 1,
        "ppca_flavor": "marginalized",
        "target_count": 10,
    }
    model = ProbabilisticPCAFromCLI(**init_args)  # type: ignore[arg-type]
    config = {
        "model": {
            "model": {
                "class_path": "cellarium.ml.models.ProbabilisticPCAFromCLI",
                "init_args": init_args,
            }
        }
    }
    module = CellariumModule(model, config=config)
    # trainer
    trainer = pl.Trainer(
        accelerator="cpu",
        devices=devices,
        max_epochs=1,
        default_root_dir=tmp_path,
    )
    # fit
    trainer.fit(module, train_dataloaders=train_loader)

    # run tests only for rank 0
    if trainer.global_rank != 0:
        return

    # load model from checkpoint
    ckpt_path = tmp_path / f"lightning_logs/version_0/checkpoints/epoch=0-step={math.ceil(n / devices)}.ckpt"
    assert ckpt_path.is_file()
    loaded_model: ProbabilisticPCAFromCLI = CellariumModule.load_from_checkpoint(ckpt_path).model
    # assert
    assert isinstance(model.transform, torch.nn.Sequential) and len(model.transform) == 2
    assert isinstance(loaded_model.transform, torch.nn.Sequential) and len(loaded_model.transform) == 2
    assert isinstance(model.transform[0], NormalizeTotal)
    assert isinstance(loaded_model.transform[0], NormalizeTotal)
    assert model.transform[0].target_count == loaded_model.transform[0].target_count
    np.testing.assert_allclose(model.W_kg.detach(), loaded_model.W_kg.detach())
    np.testing.assert_allclose(model.sigma.detach(), loaded_model.sigma.detach())
    np.testing.assert_allclose(model.mean_g.detach(), loaded_model.mean_g.detach())



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

import math
import os
from pathlib import Path

import lightning.pytorch as pl
import numpy as np
import pytest
import torch
from lightning.pytorch.strategies import DDPStrategy

from cellarium.ml import CellariumModule
from cellarium.ml.models import IncrementalPCA, IncrementalPCAFromCLI
from cellarium.ml.transforms import NormalizeTotal
from cellarium.ml.utilities.data import collate_fn
from tests.common import BoringDataset


@pytest.fixture
def x_ng():
    n, g = 10000, 100
    rng = torch.Generator()
    rng.manual_seed(1465)
    mean_g = torch.randn((g,), generator=rng)
    x_ng = torch.randn((n, g), generator=rng) + mean_g
    return x_ng


@pytest.mark.parametrize("perform_mean_correction", [False, True])
@pytest.mark.parametrize("batch_size", [10_000, 5000, 1000, 500, 250])
@pytest.mark.parametrize("k", [30, 50, 80])
def test_incremental_pca_multi_device(x_ng: torch.Tensor, perform_mean_correction: bool, batch_size: int, k: int):
    n, g = x_ng.shape
    x_ng_centered = x_ng - x_ng.mean(dim=0)
    devices = int(os.environ.get("TEST_DEVICES", "1"))
    batch_size = batch_size // devices

    # dataloader
    train_loader = torch.utils.data.DataLoader(
        BoringDataset((x_ng if perform_mean_correction else x_ng_centered).numpy()),
        batch_size=batch_size,
        shuffle=False,
        collate_fn=collate_fn,
    )
    # model
    ipca = IncrementalPCA(
        g_genes=g,
        k_components=k,
        perform_mean_correction=perform_mean_correction,
    )
    module = CellariumModule(ipca)
    # trainer
    strategy = DDPStrategy(broadcast_buffers=False) if devices > 1 else "auto"
    trainer = pl.Trainer(
        barebones=True,
        accelerator="cpu",
        devices=devices,
        max_epochs=1,
        strategy=strategy,  # type: ignore[arg-type]
    )
    # fit
    trainer.fit(module, train_dataloaders=train_loader)

    # run tests only for rank 0
    if trainer.global_rank != 0:
        return

    # actual approximation error
    x_diff = torch.linalg.matrix_norm(x_ng_centered - x_ng_centered @ ipca.V_kg.T @ ipca.V_kg, ord="fro")

    # optimal rank-k approximation error
    _, _, V_gg = torch.linalg.svd(x_ng_centered, full_matrices=False)
    V_kg = V_gg[:k]
    x_diff_rank_k = torch.linalg.matrix_norm(x_ng_centered - x_ng_centered @ V_kg.T @ V_kg, ord="fro")

    assert x_diff < x_diff_rank_k * 1.06
    assert ipca.x_size == n
    np.testing.assert_allclose(
        ipca.x_mean_g,
        (x_ng if perform_mean_correction else x_ng_centered).mean(dim=0),
        atol=1e-5,
    )


def test_load_from_checkpoint_multi_device(tmp_path: Path):
    n, g = 3, 2
    devices = int(os.environ.get("TEST_DEVICES", "1"))
    # dataloader
    train_loader = torch.utils.data.DataLoader(
        BoringDataset(np.arange(n * g).reshape(n, g)),
        collate_fn=collate_fn,
    )
    # model
    init_args = {"g_genes": g, "k_components": 1, "perform_mean_correction": True, "target_count": 10}
    model = IncrementalPCAFromCLI(**init_args)  # type: ignore[arg-type]
    config = {
        "model": {
            "model": {
                "class_path": "cellarium.ml.models.IncrementalPCAFromCLI",
                "init_args": init_args,
            }
        }
    }
    module = CellariumModule(model, config=config)
    # trainer
    strategy = DDPStrategy(broadcast_buffers=False) if devices > 1 else "auto"
    trainer = pl.Trainer(
        accelerator="cpu",
        strategy=strategy,  # type: ignore[arg-type]
        devices=devices,
        max_epochs=1,
        default_root_dir=tmp_path,
    )
    # fit
    trainer.fit(module, train_dataloaders=train_loader)

    # run tests only for rank 0
    if trainer.global_rank != 0:
        return

    # load model from checkpoint
    ckpt_path = tmp_path / f"lightning_logs/version_0/checkpoints/epoch=0-step={math.ceil(n / devices)}.ckpt"
    assert ckpt_path.is_file()
    loaded_model: IncrementalPCAFromCLI = CellariumModule.load_from_checkpoint(ckpt_path).model
    # assert
    assert isinstance(model.transform, torch.nn.Sequential) and len(model.transform) == 2
    assert isinstance(loaded_model.transform, torch.nn.Sequential) and len(loaded_model.transform) == 2
    assert isinstance(model.transform[0], NormalizeTotal)
    assert isinstance(loaded_model.transform[0], NormalizeTotal)
    assert model.transform[0].target_count == loaded_model.transform[0].target_count
    np.testing.assert_allclose(model.V_kg.detach(), loaded_model.V_kg.detach())
    np.testing.assert_allclose(model.S_k.detach(), loaded_model.S_k.detach())
    np.testing.assert_allclose(model.x_size.detach(), loaded_model.x_size.detach())
    np.testing.assert_allclose(model.x_mean_g.detach(), loaded_model.x_mean_g.detach())



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

from cellarium.ml.core.datamodule import CellariumAnnDataDataModule
from cellarium.ml.core.module import CellariumModule

__all__ = ["CellariumAnnDataDataModule", "CellariumModule"]



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

import os
from pathlib import Path
from typing import Any

import crick
import lightning.pytorch as pl
import numpy as np
import torch
import torch.distributed as dist

from cellarium.ml.models.model import CellariumModel


class TDigest(CellariumModel):
    """
    Compute an approximate non-zero histogram of the distribution of each gene in a batch of
    cells using t-digests.

    **References**:

    1. `Computing Extremely Accurate Quantiles Using T-Digests (Dunning et al.)
       <https://github.com/tdunning/t-digest/blob/master/docs/t-digest-paper/histo.pdf>`_.

    Args:
        g_genes:
            Number of genes.
        transform:
            If not ``None`` is used to transform the input data.
    """

    def __init__(self, g_genes: int, transform: torch.nn.Module | None = None) -> None:
        super().__init__()
        self.g_genes = g_genes
        self.transform = transform
        self.tdigests = [crick.tdigest.TDigest() for _ in range(g_genes)]
        self._dummy_param = torch.nn.Parameter(torch.tensor(0.0))

    @staticmethod
    def _get_fn_args_from_batch(tensor_dict: dict[str, np.ndarray | torch.Tensor]) -> tuple[tuple, dict]:
        x = tensor_dict["X"]
        return (x,), {}

    def forward(self, x_ng: torch.Tensor) -> None:
        if self.transform is not None:
            x_ng = self.transform(x_ng)

        for i, tdigest in enumerate(self.tdigests):
            x_n = x_ng[:, i]
            nonzero_mask = torch.nonzero(x_n)
            if len(nonzero_mask) > 0:
                tdigest.update(x_n[nonzero_mask])

    def on_epoch_end(self, trainer: pl.Trainer) -> None:
        # no need to merge if only one process
        if trainer.world_size == 1:
            return

        # save tdigests
        dirpath = self._resolve_ckpt_dir(trainer)
        filepath = os.path.join(dirpath, f"tdigests_{trainer.global_rank}.pt")
        trainer.strategy.checkpoint_io.save_checkpoint(self.tdigests, filepath)  # type: ignore[arg-type]
        # wait for all processes to save
        dist.barrier()

        if trainer.global_rank != 0:
            return

        # merge tdigests
        for i in range(1, trainer.world_size):  # iterate over processes
            filepath = os.path.join(dirpath, f"tdigests_{i}.pt")
            new_tdigests: list = trainer.strategy.checkpoint_io.load_checkpoint(filepath)  # type: ignore[assignment]
            for j in range(len(self.tdigests)):  # iterate over genes
                self.tdigests[j].merge(new_tdigests[j])

    @property
    def median_g(self) -> torch.Tensor:
        return torch.as_tensor([tdigest.quantile(0.5) for tdigest in self.tdigests])

    @staticmethod
    def _resolve_ckpt_dir(trainer: pl.Trainer) -> Path | str:
        if len(trainer.loggers) > 0:
            if trainer.loggers[0].save_dir is not None:
                save_dir = trainer.loggers[0].save_dir
            else:
                save_dir = trainer.default_root_dir
            name = trainer.loggers[0].name
            version = trainer.loggers[0].version
            version = version if isinstance(version, str) else f"version_{version}"
            ckpt_path = os.path.join(save_dir, str(name), version, "checkpoints")
        else:
            # if no loggers, use default_root_dir
            ckpt_path = os.path.join(trainer.default_root_dir, "checkpoints")

        os.makedirs(ckpt_path, exist_ok=True)

        return ckpt_path

    def get_extra_state(self) -> dict[str, Any]:
        return {"tdigests": self.tdigests}

    def set_extra_state(self, state: dict[str, Any]) -> None:
        self.tdigests = state["tdigests"]



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause


import torch
from torch import nn

from cellarium.ml.utilities.testing import (
    assert_nonnegative,
    assert_positive,
)


class NormalizeTotal(nn.Module):
    """
    Normalize total gene counts per cell to target count.

    .. math::

        \\mathrm{total\\_mrna\\_umis}_n = \\sum_{g=1}^G x_{ng}

        y_{ng} = \\frac{\\mathrm{target\\_count} \\times x_{ng}}{\\mathrm{total\\_mrna\\_umis}_n + \\mathrm{eps}}

    Args:
        target_count:
            Target gene epxression count.
        eps:
            A value added to the denominator for numerical stability.
    """

    def __init__(
        self,
        target_count: int = 10_000,
        eps: float = 1e-6,
    ) -> None:
        super().__init__()
        assert_positive("target_count", target_count)
        self.target_count = target_count
        assert_nonnegative("eps", eps)
        self.eps = eps

    def forward(
        self,
        x_ng: torch.Tensor,
        total_mrna_umis_n: torch.Tensor | None = None,
    ) -> torch.Tensor:
        """
        Args:
            x_ng:
                Gene counts.
            total_mrna_umis_n:
                Total mRNA UMI counts per cell. If ``None``, it is computed from ``x_ng``.

        Returns:
            Gene counts normalized to target count.
        """
        if total_mrna_umis_n is None:
            total_mrna_umis_n = x_ng.sum(dim=-1)
        return self.target_count * x_ng / (total_mrna_umis_n[:, None] + self.eps)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(target_count={self.target_count}, eps={self.eps})"



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

"""
Testing utilities
-----------------

This module contains helper functions for testing.
"""

import numpy as np
import torch


def assert_positive(name: str, number: float):
    """
    Assert that a number is positive.

    Args:
        name: The name of the number.
        number: The number to check.

    Raises:
        ValueError: If the number is not positive.
    """
    if number <= 0:
        raise ValueError(f"`{name}` must be positive. Got {number}")


def assert_nonnegative(name: str, number: float):
    """
    Assert that a number is non-negative.

    Args:
        name: The name of the number.
        number: The number to check.

    Raises:
        ValueError: If the number is negative.
    """
    if number < 0:
        raise ValueError(f"`{name}` must be non-negative. Got {number}")


def assert_columns_and_array_lengths_equal(
    matrix_name: str,
    matrix: np.ndarray | torch.Tensor,
    array_name: str,
    array: np.ndarray | torch.Tensor,
):
    """
    Assert that the number of columns in a matrix matches the length of an array.

    Args:
        matrix_name: The name of the matrix.
        matrix: The matrix.
        array_name: The name of the array.
        array: The array.

    Raises:
        ValueError: If the number of columns in the matrix does not match the length of the array.
    """
    if matrix.shape[1] != len(array):
        raise ValueError(
            f"The number of `{matrix_name}` columns must match the `{array_name}` length. "
            f"Got {matrix.shape[1]} != {len(array)}"
        )


def assert_arrays_equal(
    a1_name: str,
    a1: np.ndarray,
    a2_name: str,
    a2: np.ndarray,
):
    """
    Assert that two arrays are equal.

    Args:
        a1_name: The name of the first array.
        a1: The first array.
        a2_name: The name of the second array.
        a2: The second array.

    Raises:
        ValueError: If the arrays are not equal.
    """
    if not np.array_equal(a1, a2):
        raise ValueError(f"`{a1_name}` must match `{a2_name}`. " f"Got {a1} != {a2}")



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

from cellarium.ml.data import read_h5ad_file


def test_read_h5ad_file():
    filename = "https://storage.googleapis.com/dsp-cellarium-cas-public/test-data/test_0.h5ad"
    read_h5ad_file(filename)



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

from abc import ABCMeta, abstractmethod
from collections import OrderedDict
from collections.abc import Callable
from typing import Any

import numpy as np
import torch
from pyro.nn.module import PyroParam, _unconstrain
from torch.distributions import transform_to


class CellariumModel(torch.nn.Module, metaclass=ABCMeta):
    """
    Base class for Cellarium ML compatible models.
    """

    def __init__(self) -> None:
        super(torch.nn.Module, self).__setattr__("_pyro_params", OrderedDict())
        super().__init__()

    __call__: Callable[..., torch.Tensor | None]

    @staticmethod
    @abstractmethod
    def _get_fn_args_from_batch(tensor_dict: dict[str, np.ndarray | torch.Tensor]) -> tuple[tuple, dict]:
        """
        Get forward method arguments from batch.
        """

    def __getattr__(self, name: str) -> Any:
        if "_pyro_params" in self.__dict__:
            _pyro_params = self.__dict__["_pyro_params"]
            if name in _pyro_params:
                constraint, event_dim = _pyro_params[name]
                unconstrained_value = getattr(self, name + "_unconstrained")
                constrained_value = transform_to(constraint)(unconstrained_value)
                return constrained_value

        return super().__getattr__(name)

    def __setattr__(self, name: str, value: torch.Tensor | torch.nn.Module | PyroParam) -> None:
        if isinstance(value, PyroParam):
            # Create a new PyroParam, overwriting any old value.
            try:
                delattr(self, name)
            except AttributeError:
                pass
            constrained_value, constraint, event_dim = value
            self._pyro_params[name] = constraint, event_dim
            unconstrained_value = _unconstrain(constrained_value, constraint)
            super().__setattr__(name + "_unconstrained", unconstrained_value)
            return

        super().__setattr__(name, value)

    def __delattr__(self, name: str) -> None:
        if name in self._pyro_params:
            delattr(self, name + "_unconstrained")
            del self._pyro_params[name]
        else:
            super().__delattr__(name)


class PredictMixin(metaclass=ABCMeta):
    """
    Abstract mixin class for models that can perform prediction.
    """

    @abstractmethod
    def predict(self, x_ng: torch.Tensor, **kwargs: Any) -> torch.Tensor | dict[str, torch.Tensor | None]:
        """
        Perform prediction on data tensor.

        Args:
            x_ng:
                Data tensor.
            **kwargs:
                Additional keyword arguments.

        Returns:
            Prediction tensor or dictionary of prediction tensors.
        """



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

import gc
from collections.abc import Iterable, Sequence
from contextlib import contextmanager

import pandas as pd
from anndata import AnnData
from anndata._core.index import Index, _normalize_indices
from anndata.experimental.multi_files._anncollection import (
    AnnCollection,
    AnnCollectionView,
    ConvertType,
)
from boltons.cacheutils import LRU
from braceexpand import braceexpand

from cellarium.ml.data.fileio import read_h5ad_file
from cellarium.ml.data.schema import AnnDataSchema


class getattr_mode:
    lazy = False


_GETATTR_MODE = getattr_mode()


@contextmanager
def lazy_getattr():
    """
    When in lazy getattr mode, return a :class:`LazyAnnData` object attribute instead of
    the AnnData object attribute.
    """
    try:
        _GETATTR_MODE.lazy = True
        yield
    finally:
        _GETATTR_MODE.lazy = False


class DistributedAnnDataCollectionView(AnnCollectionView):
    """
    Distributed AnnData Collection View.

    This class is a wrapper around AnnCollectionView where adatas is a list
    of :class:`LazyAnnData` objects.
    """

    def __getitem__(self, index: Index) -> DistributedAnnDataCollectionView:
        oidx, vidx = _normalize_indices(index, self.obs_names, self.var_names)
        resolved_idx = self._resolve_idx(oidx, vidx)

        return DistributedAnnDataCollectionView(self.reference, self.convert, resolved_idx)

    @property
    def obs_names(self) -> pd.Index:
        """
        Gather and return the obs_names from all AnnData objects in the collection.
        """
        indices = []
        for i, oidx in enumerate(self.adatas_oidx):
            if oidx is None:
                continue

            adata = self.adatas[i]
            indices.append(adata.obs_names[oidx])

        if len(indices) > 1:
            concat_indices = pd.concat([pd.Series(idx) for idx in indices], ignore_index=True)
            obs_names = pd.Index(concat_indices)
            obs_names = obs_names if self.reverse is None else obs_names[self.reverse]
        else:
            obs_names = indices[0]

        return obs_names


class DistributedAnnDataCollection(AnnCollection):
    r"""
    Distributed AnnData Collection.

    This class is a wrapper around AnnCollection where adatas is a list
    of LazyAnnData objects.

    Underlying anndata files must conform to the same schema
    (see :class:`~cellarium.ml.data.schema.AnnDataSchema.validate_anndata`).
    The schema is inferred from the first AnnData file in the collection. Individual AnnData files may
    otherwise vary in the number of cells, and the actual content stored in :attr:`X`, :attr:`layers`,
    :attr:`obs` and :attr:`obsm`.

    Example 1::

        >>> dadc = DistributedAnnDataCollection(
        ...     "gs://bucket-name/folder/adata{000..005}.h5ad",
        ...     shard_size=10000,  # use if shards are sized evenly
        ...     max_cache_size=2)

    Example 2::

        >>> dadc = DistributedAnnDataCollection(
        ...     "gs://bucket-name/folder/adata{000..005}.h5ad",
        ...     shard_size=10000,
        ...     last_shard_size=6000,  # use if the size of the last shard is different
        ...     max_cache_size=2)

    Example 3::

        >>> dadc = DistributedAnnDataCollection(
        ...     "gs://bucket-name/folder/adata{000..005}.h5ad",
        ...     limits=[500, 1000, 2000, 2500, 3000, 4000],  # use if shards are sized unevenly
        ...     max_cache_size=2)

    Args:
        filenames:
            Names of anndata files.
        limits:
            List of global cell indices (limits) for the last cells in each shard.
            If ``None``, the limits are inferred from ``shard_size`` and ``last_shard_size``.
        shard_size:
            The number of cells in each anndata file (shard).
            Must be specified if the ``limits`` is not provided.
        last_shard_size:
            Last shard size. If not ``None``, the last shard will have this size possibly
            different from ``shard_size``.
        max_cache_size:
            Max size of the cache.
        cache_size_strictly_enforced:
            Assert that the number of retrieved anndatas is not more than maxsize.
        label:
            Column in :attr:`obs` to place batch information in. If it's ``None``, no column is added.
        keys:
            Names for each object being added. These values are used for column values for
            ``label`` or appended to the index if ``index_unique`` is not ``None``.
            If ``None``, ``keys`` are set to ``filenames``.
        index_unique:
            Whether to make the index unique by using the keys. If provided, this
            is the delimeter between ``{orig_idx}{index_unique}{key}``. When ``None``,
            the original indices are kept.
        convert:
            You can pass a function or a Mapping of functions which will be applied
            to the values of attributes (:attr:`obs`, :attr:`obsm`, :attr:`layers`, :attr:`X`) or to specific
            keys of these attributes in the subset object.
            Specify an attribute and a key (if needed) as keys of the passed Mapping
            and a function to be applied as a value.
        indices_strict:
            If  ``True``, arrays from the subset objects will always have the same order
            of indices as in selection used to subset.
            This parameter can be set to ``False`` if the order in the returned arrays
            is not important, for example, when using them for stochastic gradient descent.
            In this case the performance of subsetting can be a bit better.
        obs_columns:
            Subset of columns to validate in the :attr:`obs` attribute.
            If ``None``, all columns are validated.
    """

    def __init__(
        self,
        filenames: Sequence[str] | str,
        limits: Iterable[int] | None = None,
        shard_size: int | None = None,
        last_shard_size: int | None = None,
        max_cache_size: int = 1,
        cache_size_strictly_enforced: bool = True,
        label: str | None = None,
        keys: Sequence[str] | None = None,
        index_unique: str | None = None,
        convert: ConvertType | None = None,
        indices_strict: bool = True,
        obs_columns: Sequence | None = None,
    ):
        self.filenames = list(braceexpand(filenames) if isinstance(filenames, str) else filenames)
        if (shard_size is None) and (last_shard_size is not None):
            raise ValueError("If `last_shard_size` is specified then `shard_size` must also be specified.")
        if limits is None:
            assert shard_size is not None, "If `limits` is `None` then `shard_size` must be specified`"
            limits = [shard_size * (i + 1) for i in range(len(self.filenames))]
            if last_shard_size is not None:
                limits[-1] = limits[-1] - shard_size + last_shard_size
        else:
            limits = list(limits)
        assert len(limits) == len(self.filenames)
        # lru cache
        self.cache = LRU(max_cache_size)
        self.max_cache_size = max_cache_size
        self.cache_size_strictly_enforced = cache_size_strictly_enforced
        # schema
        adata0 = self.cache[self.filenames[0]] = read_h5ad_file(self.filenames[0])
        assert len(adata0) == limits[0]
        self.schema = AnnDataSchema(adata0, obs_columns)
        # lazy anndatas
        lazy_adatas = [
            LazyAnnData(filename, (start, end), self.schema, self.cache)
            for start, end, filename in zip([0] + limits, limits, self.filenames)
        ]
        # use filenames as default keys
        if keys is None:
            keys = self.filenames
        assert len(keys) == len(self.filenames)
        with lazy_getattr():
            super().__init__(
                adatas=lazy_adatas,
                join_obs=None,
                join_obsm=None,
                join_vars=None,
                label=label,
                keys=keys,
                index_unique=index_unique,
                convert=convert,
                harmonize_dtypes=False,
                indices_strict=indices_strict,
            )

    def __getitem__(self, index: Index) -> DistributedAnnDataCollectionView:
        """
        Return a distributed view of the collection of anndatas.

        :class:`LazyAnnData` instances corresponding to cells in the index are materialized.
        """
        oidx, vidx = _normalize_indices(index, self.obs_names, self.var_names)
        resolved_idx = self._resolve_idx(oidx, vidx)
        adatas_indices = [i for i, e in enumerate(resolved_idx[0]) if e is not None]
        # TODO: materialize at the last moment?
        self.materialize(adatas_indices)

        return DistributedAnnDataCollectionView(self, self.convert, resolved_idx)

    def materialize(self, indices: int | Sequence[int]) -> list[AnnData]:
        """
        Buffer and return anndata files at given indices from the list of lazy anndatas.

        This efficiently first retrieves cached files and only then caches new files.
        """
        if isinstance(indices, int):
            indices = (indices,)
        if self.cache_size_strictly_enforced:
            assert len(indices) <= self.max_cache_size, (
                f"Expected the number of anndata files ({len(indices)}) to be "
                f"no more than the max cache size ({self.max_cache_size})."
            )
        adatas = [None] * len(indices)
        # first fetch cached anndata files
        # this ensures that they are not popped if they were lru
        for i, idx in enumerate(indices):
            if self.adatas[idx].cached:
                adatas[i] = self.adatas[idx].adata
        # only then cache new anndata files
        for i, idx in enumerate(indices):
            if not self.adatas[idx].cached:
                adatas[i] = self.adatas[idx].adata
        return adatas

    def __repr__(self) -> str:
        n_obs, n_vars = self.shape
        descr = f"DistributedAnnDataCollection object with n_obs × n_vars = {self.n_obs} × {self.n_vars}"
        descr += f"\n  constructed from {len(self.filenames)} AnnData objects"
        for attr, keys in self._view_attrs_keys.items():
            if len(keys) > 0:
                descr += f"\n    view of {attr}: {str(keys)[1:-1]}"
        for attr in self._attrs:
            keys = list(getattr(self, attr).keys())
            if len(keys) > 0:
                descr += f"\n    {attr}: {str(keys)[1:-1]}"
        if "obs" in self._view_attrs_keys:
            keys = list(self.obs.keys())
            if len(keys) > 0:
                descr += f"\n    own obs: {str(keys)[1:-1]}"

        return descr

    def __getstate__(self):
        state = self.__dict__.copy()
        del state["cache"]
        del state["adatas"]
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.cache = LRU(self.max_cache_size)
        self.adatas = [
            LazyAnnData(filename, (start, end), self.schema, self.cache)
            for start, end, filename in zip([0] + self.limits, self.limits, self.filenames)
        ]


class LazyAnnData:
    """
    Lazy :class:`~anndata.AnnData` backed by a file.

    Accessing attributes under :func:`lazy_getattr` context returns schema attributes.

    Args:
        filename:
            Name of anndata file.
        limits:
            Limits of cell indices (inclusive, exclusive).
        schema:
            Schema used as a reference for lazy attributes.
        cache:
            Shared LRU cache storing buffered anndatas.
    """

    _lazy_attrs = ["obs", "obsm", "layers", "var", "varm", "varp", "var_names"]
    _all_attrs = [
        "obs",
        "var",
        "uns",
        "obsm",
        "varm",
        "layers",
        "obsp",
        "varp",
    ]

    def __init__(
        self,
        filename: str,
        limits: tuple[int, int],
        schema: AnnDataSchema,
        cache: LRU | None = None,
    ):
        self.filename = filename
        self.limits = limits
        self.schema = schema
        if cache is None:
            cache = LRU()
        self.cache = cache

    @property
    def n_obs(self) -> int:
        """Number of observations."""
        return self.limits[1] - self.limits[0]

    @property
    def n_vars(self) -> int:
        """Number of variables/features."""
        return len(self.schema.attr_values["var_names"])

    @property
    def shape(self) -> tuple[int, int]:
        """Shape of the data matrix."""
        return self.n_obs, self.n_vars

    @property
    def obs_names(self) -> pd.Index:
        """Return the observation names."""
        if _GETATTR_MODE.lazy:
            # This is only used during the initialization of DistributedAnnDataCollection
            return pd.Index([f"cell_{i}" for i in range(*self.limits)])
        else:
            return self.adata.obs_names

    @property
    def cached(self) -> bool:
        """Return whether the anndata is cached."""
        return self.filename in self.cache

    @property
    def adata(self) -> AnnData:
        """Return backed anndata from the filename"""
        try:
            adata = self.cache[self.filename]
        except KeyError:
            # fetch anndata
            adata = read_h5ad_file(self.filename)
            # validate anndata
            assert self.n_obs == adata.n_obs, (
                "Expected n_obs for LazyAnnData object and backed anndata to match "
                f"but found {self.n_obs} and {adata.n_obs}, respectively."
            )
            self.schema.validate_anndata(adata)
            # cache anndata
            if len(self.cache) < self.cache.max_size:
                self.cache[self.filename] = adata
            else:
                self.cache[self.filename] = adata
                # garbage collection of AnnData is not reliable
                # therefore we call garbage collection manually to free up the memory
                # https://github.com/scverse/anndata/issues/360
                gc.collect()
        return adata

    def __getattr__(self, attr):
        if _GETATTR_MODE.lazy:
            # This is only used during the initialization of DistributedAnnDataCollection
            if attr in self._lazy_attrs:
                return self.schema.attr_values[attr]
            raise AttributeError(f"Lazy AnnData object has no attribute '{attr}'")
        else:
            adata = self.adata
            if hasattr(adata, attr):
                return getattr(adata, attr)
            raise AttributeError(f"Backed AnnData object has no attribute '{attr}'")

    def __getitem__(self, idx) -> AnnData:
        return self.adata[idx]

    def __repr__(self) -> str:
        if self.cached:
            buffered = "Cached "
        else:
            buffered = ""
        backed_at = f" backed at {str(self.filename)!r}"
        descr = f"{buffered}LazyAnnData object with n_obs × n_vars = {self.n_obs} × {self.n_vars}{backed_at}"
        if self.cached:
            for attr in self._all_attrs:
                keys = getattr(self, attr).keys()
                if len(keys) > 0:
                    descr += f"\n    {attr}: {str(list(keys))[1:-1]}"
        return descr



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

import math
from collections.abc import Sequence

import torch
from transformers import BertConfig, BertForMaskedLM

from cellarium.ml.models.geneformer import Geneformer
from cellarium.ml.models.incremental_pca import IncrementalPCA
from cellarium.ml.models.onepass_mean_var_std import OnePassMeanVarStd
from cellarium.ml.models.probabilistic_pca import ProbabilisticPCA
from cellarium.ml.models.tdigest import TDigest
from cellarium.ml.transforms import DivideByScale, Log1p, NormalizeTotal


class OnePassMeanVarStdFromCLI(OnePassMeanVarStd):
    """
    Preset default values for the LightningCLI.

    Args:
        g_genes:
            Number of genes.
        target_count:
            Target gene expression count.
    """

    def __init__(self, g_genes: int, target_count: int = 10_000) -> None:
        transform = torch.nn.Sequential(NormalizeTotal(target_count), Log1p())
        super().__init__(g_genes, transform=transform)


class ProbabilisticPCAFromCLI(ProbabilisticPCA):
    """
    Preset default values for the LightningCLI.

    Args:
        n_cells:
            Number of cells.
        g_genes:
            Number of genes.
        k_components:
            Number of principcal components.
        ppca_flavor:
            Type of the PPCA model. Has to be one of ``marginalized`` or ``linear_vae``.
        W_init_variance_ratio:
            Ratio of variance of W_init_scale to variance of data.
            If ``mean_var_std_ckpt_path`` is ``None``, then ``W_init_scale`` is set to
            ``W_init_variance_ratio``.
        sigma_init_variance_ratio:
            Ratio of variance of sigma_init_scale to variance of data.
            If ``mean_var_std_ckpt_path`` is ``None``, then ``sigma_init_scale`` is set to
            ``sigma_init_variance_ratio``.
        seed:
            Random seed used to initialize parameters.
        target_count:
            Target gene expression count.
        mean_var_std_ckpt_path:
            Path to checkpoint containing OnePassMeanVarStd.
    """

    def __init__(
        self,
        n_cells: int,
        g_genes: int,
        k_components: int = 256,
        ppca_flavor: str = "marginalized",
        W_init_variance_ratio: float = 0.5,
        sigma_init_variance_ratio: float = 0.5,
        seed: int = 0,
        target_count: int = 10_000,
        mean_var_std_ckpt_path: str | None = None,
    ) -> None:
        if mean_var_std_ckpt_path is None:
            # compute W_init_scale and sigma_init_scale
            W_init_scale = W_init_variance_ratio
            sigma_init_scale = sigma_init_variance_ratio
            mean_g = None
        else:
            # load OnePassMeanVarStdFromCLI from checkpoint
            onepass: OnePassMeanVarStdFromCLI = torch.load(mean_var_std_ckpt_path)
            assert isinstance(onepass.transform, torch.nn.Sequential)
            assert isinstance(onepass.transform[0], NormalizeTotal)
            assert target_count == onepass.transform.target_count
            assert g_genes == onepass.g_genes
            # compute W_init_scale and sigma_init_scale
            total_variance = onepass.var_g.sum().item()
            W_init_scale = math.sqrt(W_init_variance_ratio * total_variance / (g_genes * k_components))
            sigma_init_scale = math.sqrt(sigma_init_variance_ratio * total_variance / g_genes)
            mean_g = onepass.mean_g
        # create transform
        transform = torch.nn.Sequential(NormalizeTotal(target_count), Log1p())
        self.mean_var_std_ckpt_path = mean_var_std_ckpt_path
        super().__init__(
            n_cells=n_cells,
            g_genes=g_genes,
            k_components=k_components,
            ppca_flavor=ppca_flavor,
            mean_g=mean_g,
            W_init_scale=W_init_scale,
            sigma_init_scale=sigma_init_scale,
            seed=seed,
            transform=transform,
        )


class IncrementalPCAFromCLI(IncrementalPCA):
    """
    Preset default values for the LightningCLI.

    Args:
        k_components:
            Number of principal components.
        svd_lowrank_niter:
            Number of iterations for the low-rank SVD algorithm.
        perform_mean_correction:
            If ``True`` then the mean correction is applied to the update step.
            If ``False`` then the data is assumed to be centered and the mean correction
            is not applied to the update step.
        target_count:
            Target gene epxression count.
    """

    def __init__(
        self,
        g_genes: int,
        k_components: int,
        svd_lowrank_niter: int = 2,
        perform_mean_correction: bool = False,
        target_count: int = 10_000,
    ) -> None:
        transform = torch.nn.Sequential(NormalizeTotal(target_count), Log1p())
        super().__init__(
            g_genes=g_genes,
            k_components=k_components,
            svd_lowrank_niter=svd_lowrank_niter,
            perform_mean_correction=perform_mean_correction,
            transform=transform,
        )


class TDigestFromCLI(TDigest):
    """
    Preset default values for the LightningCLI.

    Args:
        g_genes:
            Number of genes.
        target_count:
            Target gene epxression count.
        eps:
            A value added to the denominator for numerical stability.
    """

    def __init__(self, g_genes: int, target_count: int = 10_000, eps: float = 1e-6) -> None:
        transform = NormalizeTotal(target_count=target_count, eps=eps)
        super().__init__(g_genes, transform=transform)


class GeneformerFromCLI(Geneformer):
    """
    Preset default values for the LightningCLI.

    Args:
        feature_schema:
            The list of the variable names in the input data.
        hidden_size:
            Dimensionality of the encoder layers and the pooler layer.
        num_hidden_layers:
            Number of hidden layers in the Transformer encoder.
        num_attention_heads:
            Number of attention heads for each attention layer in the Transformer encoder.
        intermediate_size:
            Dimensionality of the "intermediate" (often named feed-forward) layer in the Transformer encoder.
        hidden_act:
            The non-linear activation function (function or string) in the encoder and pooler. If string, ``"gelu"``,
            ``"relu"``, ``"silu"`` and ``"gelu_new"`` are supported.
        hidden_dropout_prob:
            The dropout probability for all fully connected layers in the embeddings, encoder, and pooler.
        attention_probs_dropout_prob:
            The dropout ratio for the attention probabilities.
        max_position_embeddings:
            The maximum sequence length that this model might ever be used with. Typically set this to something large
            just in case (e.g., 512 or 1024 or 2048).
        type_vocab_size:
            The vocabulary size of the ``token_type_ids`` passed when calling :class:`transformers.BertModel`.
        initializer_range:
            The standard deviation of the truncated_normal_initializer for initializing all weight matrices.
        position_embedding_type:
            Type of position embedding. Choose one of ``"absolute"``, ``"relative_key"``, ``"relative_key_query"``. For
            positional embeddings use ``"absolute"``. For more information on ``"relative_key"``, please refer to
            `Self-Attention with Relative Position Representations (Shaw et al.) <https://arxiv.org/abs/1803.02155>`_.
            For more information on ``"relative_key_query"``, please refer to *Method 4* in `Improve Transformer Models
            with Better Relative Position Embeddings (Huang et al.) <https://arxiv.org/abs/2009.13658>`_.
        layer_norm_eps:
            The epsilon used by the layer normalization layers.
        mlm_probability:
            Ratio of tokens to mask for masked language modeling loss.
        tdigest_path:
            Path to the tdigest checkpoint. The tdigest checkpoint is used to normalize the input data by the non-zero
            median gene count values. If ``None`` then no normalization is applied.
        validate_input:
            If ``True`` the input data is validated.
    """

    def __init__(
        self,
        feature_schema: Sequence,
        hidden_size: int = 256,
        num_hidden_layers: int = 6,
        num_attention_heads: int = 4,
        intermediate_size: int = 512,
        hidden_act: str = "relu",
        hidden_dropout_prob: float = 0.02,
        attention_probs_dropout_prob: float = 0.02,
        max_position_embeddings: int = 2048,
        type_vocab_size: int = 2,
        initializer_range: float = 0.02,
        position_embedding_type: str = "absolute",
        layer_norm_eps: float = 1e-12,
        mlm_probability: float = 0.15,
        tdigest_path: str | None = None,
        validate_input: bool = True,
    ):
        # model configuration
        config = {
            "vocab_size": len(feature_schema) + 2,  # number of genes + 2 for <mask> and <pad> tokens
            "hidden_size": hidden_size,
            "num_hidden_layers": num_hidden_layers,
            "num_attention_heads": num_attention_heads,
            "intermediate_size": intermediate_size,
            "hidden_act": hidden_act,
            "attention_probs_dropout_prob": attention_probs_dropout_prob,
            "hidden_dropout_prob": hidden_dropout_prob,
            "max_position_embeddings": max_position_embeddings,
            "type_vocab_size": type_vocab_size,
            "initializer_range": initializer_range,
            "position_embedding_type": position_embedding_type,
            "layer_norm_eps": layer_norm_eps,
            "pad_token_id": 0,
        }
        config = BertConfig(**config)
        model = BertForMaskedLM(config)

        if tdigest_path is not None:
            tdigest = torch.load(tdigest_path)
            transform = torch.nn.Sequential(
                NormalizeTotal(
                    target_count=tdigest.transform.target_count,
                    eps=tdigest.transform.eps,
                ),
                DivideByScale(
                    scale_g=tdigest.median_g,
                    feature_schema=feature_schema,
                    eps=tdigest.transform.eps,
                ),
            )
        else:
            transform = None
        super().__init__(
            feature_schema=feature_schema,
            model=model,
            mlm_probability=mlm_probability,
            transform=transform,
            validate_input=validate_input,
        )



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

from collections.abc import Sequence
from typing import Any

import numpy as np
import torch
from transformers import BertForMaskedLM

from cellarium.ml.models.model import CellariumModel, PredictMixin


class Geneformer(CellariumModel, PredictMixin):
    """
    Geneformer model.

    **References:**

    1. `Transfer learning enables predictions in network biology (Theodoris et al.)
       <https://www.nature.com/articles/s41586-023-06139-9>`_.

    Args:
        feature_schema:
            The list of the variable names in the input data.
        model:
            The bert model.
        mlm_probability:
            Ratio of tokens to mask for masked language modeling loss.
        transform:
            If not ``None`` is used to transform the input data.
        validate_input:
            If ``True`` the input data is validated.
    """

    def __init__(
        self,
        feature_schema: Sequence,
        model: BertForMaskedLM,
        mlm_probability: float,
        transform: torch.nn.Module | None = None,
        validate_input: bool = True,
    ):
        super().__init__()
        self.feature_schema = feature_schema
        self.model = model
        self.mlm_probability = mlm_probability
        self.transform = transform
        self.validate_input = validate_input
        self.feature_ids: torch.Tensor
        # ids for the features, 0 is for padding, 1 is for mask
        self.register_buffer("feature_ids", torch.arange(2, len(feature_schema) + 2))

    @staticmethod
    def _get_fn_args_from_batch(tensor_dict: dict[str, np.ndarray | torch.Tensor]) -> tuple[tuple, dict]:
        x = tensor_dict["X"]
        feature_list = tensor_dict["var_names"]
        return (x,), {"feature_list": feature_list}

    def tokenize(self, x_ng: torch.Tensor, feature_list: Sequence) -> tuple[torch.Tensor, torch.Tensor]:
        tokens = self.feature_ids.expand(x_ng.shape)
        # sort by median-scaled gene values
        sorted_indices = torch.argsort(x_ng, dim=1, descending=True)
        sorted_indices = sorted_indices[:, : self.model.config.max_position_embeddings]
        ndx = torch.arange(x_ng.shape[0], device=x_ng.device)
        input_ids = tokens[ndx[:, None], sorted_indices]
        # mask out genes with zero expression
        sorted_x_ng = x_ng[ndx[:, None], sorted_indices]
        attention_mask = sorted_x_ng != 0
        # pad genes with zero expression
        input_ids[~attention_mask] = 0
        return input_ids, attention_mask

    def forward(self, x_ng: torch.Tensor, **kwargs: Any) -> torch.Tensor:
        assert "feature_list" in kwargs, "feature_list must be provided."
        feature_list: Sequence = kwargs.pop("feature_list")

        if self.validate_input:
            assert x_ng.shape[1] == len(feature_list), "The number of x_ng columns must match the feature_list length."
            assert np.array_equal(feature_list, self.feature_schema), "feature_list must match the feature_schema."

        if self.transform is not None:
            x_ng = self.transform(x_ng)

        input_ids, attention_mask = self.tokenize(x_ng, feature_list)

        labels = input_ids.clone()
        labels_probs = torch.full(labels.shape, self.mlm_probability, device=x_ng.device)
        labels_probs[~attention_mask] = 0
        masked_indices = torch.bernoulli(labels_probs).bool()
        labels[~masked_indices] = -100  # we only compute loss on masked tokens

        # 80% of the time, we replace masked input tokens with tokenizer.mask_token ([MASK])
        indices_replaced = torch.bernoulli(torch.full(labels.shape, 0.8, device=x_ng.device)).bool() & masked_indices
        input_ids[indices_replaced] = 1  # tokenizer.mask_token_id

        # 10% of the time, we replace masked input tokens with random word
        indices_random = (
            torch.bernoulli(torch.full(labels.shape, 0.5, device=x_ng.device)).bool()
            & masked_indices
            & ~indices_replaced
        )
        random_words = torch.randint(x_ng.shape[1], labels.shape, dtype=torch.long, device=x_ng.device)
        input_ids[indices_random] = random_words[indices_random]

        output = self.model(
            input_ids=input_ids,
            attention_mask=attention_mask,
            labels=labels,
        )
        return output.loss

    def predict(self, x_ng: torch.Tensor, **kwargs: Any) -> dict[str, torch.Tensor | None]:
        assert "feature_list" in kwargs, "feature_list must be provided."
        feature_list: Sequence = kwargs.pop("feature_list")
        output_hidden_states: bool = kwargs.pop("output_hidden_states", True)
        output_attentions: bool = kwargs.pop("output_attentions", True)

        if self.validate_input:
            assert x_ng.shape[1] == len(feature_list), "The number of x_ng columns must match the feature_list length."
            assert np.array_equal(feature_list, self.feature_schema), "feature_list must match the feature_schema."

        if self.transform is not None:
            x_ng = self.transform(x_ng)

        input_ids, attention_mask = self.tokenize(x_ng, feature_list)

        output = self.model(
            input_ids=input_ids,
            attention_mask=attention_mask,
            output_hidden_states=output_hidden_states,
            output_attentions=output_attentions,
        )
        return output



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

from cellarium.ml.preprocessing.highly_variable_genes import get_highly_variable_genes

__all__ = ["get_highly_variable_genes"]



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

from cellarium.ml.models.from_cli import (
    GeneformerFromCLI,
    IncrementalPCAFromCLI,
    OnePassMeanVarStdFromCLI,
    ProbabilisticPCAFromCLI,
    TDigestFromCLI,
)
from cellarium.ml.models.gather import GatherLayer
from cellarium.ml.models.geneformer import Geneformer
from cellarium.ml.models.incremental_pca import IncrementalPCA
from cellarium.ml.models.logistic_regression import LogisticRegression
from cellarium.ml.models.model import CellariumModel, PredictMixin
from cellarium.ml.models.onepass_mean_var_std import OnePassMeanVarStd
from cellarium.ml.models.probabilistic_pca import ProbabilisticPCA
from cellarium.ml.models.tdigest import TDigest

__all__ = [
    "CellariumModel",
    "GatherLayer",
    "Geneformer",
    "GeneformerFromCLI",
    "IncrementalPCA",
    "IncrementalPCAFromCLI",
    "LogisticRegression",
    "OnePassMeanVarStd",
    "OnePassMeanVarStdFromCLI",
    "PredictMixin",
    "ProbabilisticPCA",
    "ProbabilisticPCAFromCLI",
    "TDigest",
    "TDigestFromCLI",
]



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

from __future__ import annotations

from importlib import import_module
from typing import IO, Any
from unittest.mock import patch

import lightning.pytorch as pl
import numpy as np
import torch
from jsonargparse import Namespace
from lightning.fabric.utilities.types import _MAP_LOCATION_TYPE, _PATH
from lightning.pytorch.utilities.types import STEP_OUTPUT, OptimizerLRSchedulerConfig

from cellarium.ml.core.saving import _load_state
from cellarium.ml.models import CellariumModel, PredictMixin


class CellariumModule(pl.LightningModule):
    """
    ``CellariumModule`` organizes code into following sections:

        * :attr:`model`: A :class:`cellarium.ml.models.CellariumModel` to train with
          :meth:`training_step` method and epoch end hooks.
        * :attr:`optim_fn` and :attr:`optim_kwargs`: A Pytorch optimizer class and its keyword arguments.
        * :attr:`scheduler_fn` and :attr:`scheduler_kwargs`: A Pytorch lr scheduler class and its
          keyword arguments.

    Args:
        model:
            A :class:`cellarium.ml.models.CellariumModel` to train.
        optim_fn:
            A Pytorch optimizer class, e.g., :class:`~torch.optim.Adam`. If ``None``,
            defaults to :class:`torch.optim.Adam`.
        optim_kwargs:
            Keyword arguments for optimiser. If ``None``, defaults to ``default_lr``.
        scheduler_fn:
            A Pytorch lr scheduler class, e.g., :class:`~torch.optim.lr_scheduler.CosineAnnealingLR`.
        scheduler_kwargs:
            Keyword arguments for lr scheduler.
        default_lr:
            Default learning rate to use if ``optim_kwargs`` does not contain ``lr``.
        config:
            A dictionary or :class:`jsonargparse.Namespace` containing the initialization hyperparameters.
            If not ``None``, the configuration will be saved as ``"hyper_parameters"`` in the checkpoint.
    """

    def __init__(
        self,
        model: CellariumModel,
        optim_fn: type[torch.optim.Optimizer] | str | None = None,
        optim_kwargs: dict | None = None,
        scheduler_fn: type[torch.optim.lr_scheduler.LRScheduler] | str | None = None,
        scheduler_kwargs: dict | None = None,
        default_lr: float = 1e-3,
        config: dict[str, Any] | Namespace | None = None,
    ) -> None:
        super().__init__()
        self.model = model

        # set up optimizer and scheduler
        if isinstance(optim_fn, str):
            class_module, class_name = optim_fn.rsplit(".", 1)
            optim_module = import_module(class_module)
            self.optim_fn = getattr(optim_module, class_name)
        elif optim_fn is None:
            self.optim_fn = torch.optim.Adam
        else:
            self.optim_fn = optim_fn

        if isinstance(scheduler_fn, str):
            class_module, class_name = scheduler_fn.rsplit(".", 1)
            scheduler_module = import_module(class_module)
            self.scheduler_fn = getattr(scheduler_module, class_name)
        else:
            self.scheduler_fn = scheduler_fn

        optim_kwargs = {} if optim_kwargs is None else optim_kwargs
        if "lr" not in optim_kwargs:
            optim_kwargs["lr"] = default_lr
        self.optim_kwargs = optim_kwargs
        self.scheduler_fn = scheduler_fn
        self.scheduler_kwargs = scheduler_kwargs

        if config is not None:
            self._set_hparams(config)

    @classmethod
    @patch("lightning.pytorch.core.saving._load_state", new=_load_state)
    def load_from_checkpoint(
        cls,
        checkpoint_path: _PATH | IO,
        map_location: _MAP_LOCATION_TYPE = None,
        hparams_file: _PATH | None = None,
        strict: bool = True,
        **kwargs: Any,
    ) -> CellariumModule:
        r"""
        Primary way of loading a model from a checkpoint. When Cellarium ML saves a checkpoint it stores the config
        argument passed to ``__init__``  in the checkpoint under ``"hyper_parameters"``.

        Any arguments specified through ``**kwargs`` will override args stored in ``"hyper_parameters"``.

        Args:
            checkpoint_path:
                Path to checkpoint. This can also be a URL, or file-like object
            map_location:
                If your checkpoint saved a GPU model and you now load on CPUs
                or a different number of GPUs, use this to map to the new setup.
                The behaviour is the same as in :func:`torch.load`.
            hparams_file:
                Optional path to a ``.yaml`` or ``.csv`` file with hierarchical structure as in this example:

                .. code-block:: yaml

                    model:
                      model:
                        class_path: cellarium.ml.models.OnePassMeanVarStdFromCLI
                        init_args:
                          g_genes: 36350
                          target_count: 10000
                      optim_fn: null
                      optim_kwargs: null
                      scheduler_fn: null
                      scheduler_kwargs: null
                      default_lr: 0.001

                If you train a model using :mod:`cellarium.ml.cli` module you most likely won't need this
                since Cellarium ML CLI will always save the hyperparameters to the checkpoint.

                However, if your checkpoint weights don't have the hyperparameters saved,
                use this method to pass in a ``.yaml`` file with the hparams you'd like to use.
                These will be converted into a :class:`dict` and passed into your
                :class:`CellariumModule` for use.
            strict:
                Whether to strictly enforce that the keys in :attr:`checkpoint_path` match the keys
                returned by this module's state dict.
            \**kwargs: Any extra keyword args needed to init the model. Can also be used to override saved
                hyperparameter values.

        Return:
            :class:`CellariumModule` instance with loaded weights and hyperparameters.

        Example::

            # load weights without mapping ...
            module = CellariumModule.load_from_checkpoint("path/to/checkpoint.ckpt")

            # or load weights mapping all weights from GPU 1 to GPU 0 ...
            map_location = {"cuda:1": "cuda:0"}
            module = CellariumModule.load_from_checkpoint(
                "path/to/checkpoint.ckpt",
                map_location=map_location
            )

            # or load weights and hyperparameters from separate files.
            module = CellariumModule.load_from_checkpoint(
                "path/to/checkpoint.ckpt",
                hparams_file="/path/to/config.yaml"
            )

            # override some of the params with new values
            module = CellariumModule.load_from_checkpoint(
                "path/to/checkpoint.ckpt",
                optim_fn=torch.optim.AdamW,
                default_lr=0.0001,
            )
        """
        return super().load_from_checkpoint(
            checkpoint_path=checkpoint_path,
            map_location=map_location,
            hparams_file=hparams_file,
            strict=strict,
            **kwargs,
        )

    def training_step(  # type: ignore[override]
        self, batch: dict[str, np.ndarray | torch.Tensor], batch_idx: int
    ) -> torch.Tensor | None:
        args, kwargs = self.model._get_fn_args_from_batch(batch)
        loss = self.model(*args, **kwargs)
        if loss is not None:
            # Logging to TensorBoard by default
            self.log("train_loss", loss)
        return loss

    def forward(self, batch: dict[str, np.ndarray | torch.Tensor]) -> torch.Tensor | dict[str, torch.Tensor | None]:
        """Forward pass of the model."""
        assert isinstance(self.model, PredictMixin)
        args, kwargs = self.model._get_fn_args_from_batch(batch)
        return self.model.predict(*args, **kwargs)

    def configure_optimizers(self) -> OptimizerLRSchedulerConfig:
        """Configure optimizers for the model."""
        optim_config: OptimizerLRSchedulerConfig = {
            "optimizer": self.optim_fn(self.model.parameters(), **self.optim_kwargs)
        }
        if self.scheduler_fn is not None:
            assert self.scheduler_kwargs is not None
            scheduler = self.scheduler_fn(optim_config["optimizer"], **self.scheduler_kwargs)
            optim_config["lr_scheduler"] = {"scheduler": scheduler, "interval": "step"}
        return optim_config

    def on_train_epoch_start(self) -> None:
        """
        Calls the ``set_epoch`` method on the iterable dataset of the given dataloader.

        If the dataset is ``IterableDataset`` and has ``set_epoch`` method defined, then
        ``set_epoch`` must be called at the beginning of every epoch to ensure shuffling
        applies a new ordering. This has no effect if shuffling is off.
        """
        # dataloader is wrapped in a combined loader and can be accessed via
        # flattened property which returns a list of dataloaders
        # https://lightning.ai/docs/pytorch/stable/api/lightning.pytorch.utilities.combined_loader.html
        combined_loader = self.trainer.fit_loop._combined_loader
        assert combined_loader is not None
        dataloaders = combined_loader.flattened
        for dataloader in dataloaders:
            dataset = dataloader.dataset
            set_epoch = getattr(dataset, "set_epoch", None)
            if callable(set_epoch):
                set_epoch(self.current_epoch)

    def on_train_start(self) -> None:
        """
        Calls the ``on_train_start`` method on the :attr:`model` attribute.
        If the :attr:`model` attribute has ``on_train_start`` method defined, then
        ``on_train_start`` must be called at the beginning of training.
        """
        on_train_start = getattr(self.model, "on_train_start", None)
        if callable(on_train_start):
            on_train_start(self.trainer)

    def on_train_epoch_end(self) -> None:
        """
        Calls the ``on_epoch_end`` method on the :attr:`model` attribute.
        If the :attr:`model` attribute has ``on_epoch_end`` method defined, then
        ``on_epoch_end`` must be called at the end of every epoch.
        """
        on_epoch_end = getattr(self.model, "on_epoch_end", None)
        if callable(on_epoch_end):
            on_epoch_end(self.trainer)

    def on_train_batch_end(self, outputs: STEP_OUTPUT, batch: Any, batch_idx: int) -> None:
        """
        Calls the ``on_batch_end`` method on the module.
        """
        on_batch_end = getattr(self.model, "on_batch_end", None)
        if callable(on_batch_end):
            on_batch_end(self.trainer)



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

import math
from typing import Any

import lightning.pytorch as pl
import numpy as np
import torch
import torch.distributed as dist
import torch.nn as nn
from lightning.pytorch.strategies import DDPStrategy

from cellarium.ml.models.model import CellariumModel, PredictMixin


class IncrementalPCA(CellariumModel, PredictMixin):
    """
    Distributed and Incremental PCA.

    **References:**

    1. `A Distributed and Incremental SVD Algorithm for Agglomerative Data Analysis on Large Networks (Iwen et al.)
       <https://users.math.msu.edu/users/iwenmark/Papers/distrib_inc_svd.pdf>`_.
    2. `Incremental Learning for Robust Visual Tracking (Ross et al.)
       <https://www.cs.toronto.edu/~dross/ivt/RossLimLinYang_ijcv.pdf>`_.

    Args:
        g_genes:
            Number of genes.
        k_components:
            Number of principal components.
        svd_lowrank_niter:
            Number of iterations for the low-rank SVD algorithm.
        perform_mean_correction:
            If ``True`` then the mean correction is applied to the update step.
            If ``False`` then the data is assumed to be centered and the mean correction
            is not applied to the update step.
        transform:
            If not ``None`` is used to transform the input data.
    """

    def __init__(
        self,
        g_genes: int,
        k_components: int,
        svd_lowrank_niter: int = 2,
        perform_mean_correction: bool = False,
        transform: nn.Module | None = None,
    ) -> None:
        super().__init__()
        self.g_genes = g_genes
        self.k_components = k_components
        self.svd_lowrank_niter = svd_lowrank_niter
        self.perform_mean_correction = perform_mean_correction
        self.transform = transform
        self.V_kg: torch.Tensor
        self.S_k: torch.Tensor
        self.x_mean_g: torch.Tensor
        self.x_size: torch.Tensor
        self.register_buffer("V_kg", torch.zeros(k_components, g_genes))
        self.register_buffer("S_k", torch.zeros(k_components))
        self.register_buffer("x_mean_g", torch.zeros(g_genes))
        self.register_buffer("x_size", torch.tensor(0))
        self._dummy_param = nn.Parameter(torch.tensor(0.0))

    @staticmethod
    def _get_fn_args_from_batch(tensor_dict: dict[str, np.ndarray | torch.Tensor]) -> tuple[tuple, dict]:
        x = tensor_dict["X"]
        return (x,), {}

    def forward(self, x_ng: torch.Tensor) -> None:
        """
        Incrementally update partial SVD with new data.
        """
        if self.transform is not None:
            x_ng = self.transform(x_ng)

        g = self.g_genes
        k = self.k_components
        niter = self.svd_lowrank_niter
        self_X_size = self.x_size
        other_X_size = x_ng.size(0)
        total_X_size = self_X_size + other_X_size
        assert k <= min(other_X_size, g), (
            f"Rank of svd_lowrank (k_components): {k}"
            f" must be less than min(n_cells, g_genes): {min(other_X_size, g)}"
        )

        # compute SVD of new data
        if self.perform_mean_correction:
            self_X_mean = self.x_mean_g
            other_X_mean = x_ng.mean(dim=0)
            _, S_k, V_gk = torch.svd_lowrank(x_ng - other_X_mean, q=k, niter=niter)
        else:
            _, S_k, V_gk = torch.svd_lowrank(x_ng, q=k, niter=niter)

        # if not the first batch, merge results
        if self_X_size > 0:
            self_X = torch.einsum("k,kg->kg", self.S_k, self.V_kg)
            other_X = torch.einsum("k,gk->kg", S_k, V_gk)
            joined_X = torch.cat([self_X, other_X], dim=0)
            if self.perform_mean_correction:
                mean_correction = (
                    math.sqrt(self_X_size * other_X_size / total_X_size) * (self_X_mean - other_X_mean)[None, :]
                )
                joined_X = torch.cat([joined_X, mean_correction], dim=0)
            # perform SVD on merged results
            _, S_k, V_gk = torch.svd_lowrank(joined_X, q=k, niter=niter)

        # update buffers
        self.V_kg = V_gk.T
        self.S_k = S_k
        self.x_size = total_X_size
        if self.perform_mean_correction:
            self.x_mean_g = self_X_mean * self_X_size / total_X_size + other_X_mean * other_X_size / total_X_size

    def on_train_start(self, trainer: pl.Trainer) -> None:
        if trainer.world_size > 1:
            assert isinstance(trainer.strategy, DDPStrategy), (
                "Distributed and Incremental PCA requires that " "the trainer uses the DDP strategy."
            )
            assert trainer.strategy._ddp_kwargs["broadcast_buffers"] is False, (
                "Distributed and Incremental PCA requires that " "broadcast_buffers is set to False."
            )

    def on_epoch_end(self, trainer: pl.Trainer) -> None:
        """
        Merge partial SVD results from parallel processes at the end of the epoch.

        Merging SVDs is performed hierarchically. At each merging level, the leading
        process (even rank) merges its SVD with the trailing process (odd rank).
        The trailing process discards its SVD and is terminated. The leading
        process continues to the next level. This process continues until only
        one process remains. The final SVD is stored on the remaining process.

        The number of levels (hierarchy depth) scales logarithmically with the
        number of processes.
        """
        # no need to merge if only one process
        if trainer.world_size == 1:
            return

        assert trainer.current_epoch == 0, "Only one pass through the data is required."

        # parameters
        k = self.k_components
        niter = self.svd_lowrank_niter
        self_X_size = self.x_size
        self_X = torch.einsum("k,kg->kg", self.S_k, self.V_kg).contiguous()
        if self.perform_mean_correction:
            self_X_mean = self.x_mean_g

        # initialize rank, world_size, and merging level i
        rank = trainer.global_rank
        world_size = trainer.world_size
        i = 0
        while world_size > 1:
            # level i
            # at most two local ranks (leading and trailing)
            if rank % 2 == 0:  # leading rank
                if rank < world_size:
                    # if there is a trailing rank
                    # then concatenate and merge SVDs
                    src = (rank + 1) * 2**i
                    other_X = torch.zeros_like(self_X)
                    other_X_size = torch.zeros_like(self_X_size)
                    dist.recv(other_X, src=src)
                    dist.recv(other_X_size, src=src)
                    total_X_size = self_X_size + other_X_size

                    # obtain joined_X
                    joined_X = torch.cat([self_X, other_X], dim=0)
                    if self.perform_mean_correction:
                        other_X_mean = torch.zeros_like(self_X_mean)
                        dist.recv(other_X_mean, src=src)
                        mean_correction = (
                            math.sqrt(self_X_size * other_X_size / total_X_size) * (self_X_mean - other_X_mean)[None, :]
                        )
                        joined_X = torch.cat([joined_X, mean_correction], dim=0)

                    # perform SVD on joined_X
                    _, S_k, V_gk = torch.svd_lowrank(joined_X, q=k, niter=niter)

                    # update parameters
                    V_kg = V_gk.T
                    if self.perform_mean_correction:
                        self_X_mean = (
                            self_X_mean * self_X_size / total_X_size + other_X_mean * other_X_size / total_X_size
                        )
                    self_X = torch.einsum("k,kg->kg", S_k, V_kg).contiguous()
                    self_X_size = total_X_size
            else:  # trailing rank
                # send to a leading rank and exit
                dst = (rank - 1) * 2**i
                dist.send(self_X, dst=dst)
                dist.send(self_X_size, dst=dst)
                if self.perform_mean_correction:
                    dist.send(self_X_mean, dst=dst)
                break
            # update rank, world_size, and level i
            rank = rank // 2
            world_size = math.ceil(world_size / 2)
            i += 1
        else:
            assert trainer.global_rank == 0
            self.V_kg = V_kg
            self.S_k = S_k
            self.x_size = self_X_size
            if self.perform_mean_correction:
                self.x_mean_g = self_X_mean

    @property
    def explained_variance_k(self) -> torch.Tensor:
        r"""
        The amount of variance explained by each of the selected components. The variance
        estimation uses ``x_size`` degrees of freedom.

        Equal to ``k_components`` largest eigenvalues of the covariance matrix of input data.
        """
        return self.S_k**2 / self.x_size

    @property
    def components_kg(self) -> torch.Tensor:
        r"""
        Principal axes in feature space, representing the directions of maximum variance
        in the data. Equivalently, the right singular vectors of the centered input data,
        parallel to its eigenvectors. The components are sorted by decreasing ``explained_variance_k``.
        """
        return self.V_kg

    def predict(self, x_ng: torch.Tensor, **kwargs: Any) -> torch.Tensor:
        r"""
        Centering and embedding of the input data ``x_ng`` into the principal component space.
        """
        if self.transform is not None:
            x_ng = self.transform(x_ng)
        return (x_ng - self.x_mean_g) @ self.V_kg.T



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

from functools import cache
from typing import Any

import numpy as np
import torch
from numpy.typing import ArrayLike
from torch import nn

from cellarium.ml.utilities.testing import (
    assert_columns_and_array_lengths_equal,
)


class Filter(nn.Module):
    """
    Filter gene counts by a list of features.

    .. math::

        \\mathrm{mask}_g = \\mathrm{feature}_g \\in \\mathrm{filter\\_list}

        y_{ng} = x_{ng}[:, \\mathrm{mask}_g]

    Args:
        filter_list: A list of features to filter by.
    """

    def __init__(self, filter_list: ArrayLike) -> None:
        super().__init__()
        self.filter_list = np.array(filter_list)
        if len(self.filter_list) == 0:
            raise ValueError(f"`filter_list` must not be empty. Got {self.filter_list}")

    @cache
    def filter(self, feature_g: tuple) -> np.ndarray[Any, np.dtype[np.bool_]]:
        """
        Args:
            feature_g: The list of the variable names in the input data.

        Returns:
            A boolean mask of the features to filter by.
        """
        mask = np.isin(feature_g, self.filter_list)
        if not np.any(mask):
            raise AssertionError("No features in `feature_g` matched the `filter_list`")
        return mask

    def forward(self, x_ng: torch.Tensor, feature_g: np.ndarray) -> torch.Tensor:
        """
        Args:
            x_ng:
                Gene counts.
            feature_g:
                The list of the variable names in the input data.

        Returns:
            Filtered gene counts.
        """
        assert_columns_and_array_lengths_equal("x_ng", x_ng, "feature_g", feature_g)

        filter_mask = self.filter(tuple(feature_g.tolist()))
        return x_ng[:, filter_mask]

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(filter_list={self.filter_list})"



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

import math
import os
import shutil
from pathlib import Path

import lightning.pytorch as pl
import numpy as np
import pytest
import torch
from anndata import AnnData

from cellarium.ml import CellariumModule
from cellarium.ml.data import (
    DistributedAnnDataCollection,
    IterableDistributedAnnDataCollectionDataset,
)
from cellarium.ml.models import TDigest, TDigestFromCLI
from cellarium.ml.transforms import NormalizeTotal
from cellarium.ml.utilities.data import AnnDataField, collate_fn
from tests.common import BoringDataset


@pytest.fixture
def adata():
    n_cell, g_gene = 1000, 5
    rng = np.random.default_rng(1465)
    X = rng.integers(100, size=(n_cell, g_gene))
    return AnnData(X, dtype=X.dtype)


@pytest.fixture
def dadc(adata: AnnData, tmp_path: Path):
    # save anndata files
    limits = [200, 500, 1000]
    for i, limit in enumerate(zip([0] + limits, limits)):
        sliced_adata = adata[slice(*limit)]
        sliced_adata.write(os.path.join(tmp_path, f"adata.00{i}.h5ad"))

    # distributed anndata
    filenames = str(os.path.join(tmp_path, "adata.{000..002}.h5ad"))
    return DistributedAnnDataCollection(
        filenames,
        limits,
        max_cache_size=2,  # allow max_cache_size=2 for IterableDistributedAnnDataCollectionDataset
        cache_size_strictly_enforced=True,
    )


@pytest.mark.parametrize("shuffle", [False, True])
@pytest.mark.parametrize("num_workers", [0, 2])
@pytest.mark.parametrize("batch_size", [10, 100])
@pytest.mark.parametrize(
    "transform",
    [NormalizeTotal(target_count=10_000, eps=0), None],
)
def test_tdigest_multi_device(
    adata: AnnData,
    dadc: DistributedAnnDataCollection,
    shuffle: bool,
    num_workers: int,
    batch_size: int,
    transform: torch.nn.Module | None,
):
    devices = int(os.environ.get("TEST_DEVICES", "1"))
    batch_size = batch_size // devices

    # prepare dataloader
    dataset = IterableDistributedAnnDataCollectionDataset(
        dadc,
        batch_keys={"X": AnnDataField("X")},
        batch_size=batch_size,
        shuffle=shuffle,
    )
    data_loader = torch.utils.data.DataLoader(
        dataset,
        num_workers=num_workers,
        collate_fn=collate_fn,
    )

    # fit
    model = TDigest(g_genes=dadc.n_vars, transform=transform)
    module = CellariumModule(model)
    trainer = pl.Trainer(
        barebones=True,
        accelerator="cpu",
        devices=devices,
        max_epochs=1,  # one pass
    )
    trainer.fit(module, train_dataloaders=data_loader)

    # run tests only for rank 0
    if trainer.global_rank != 0:
        return

    # actual median
    actual_median_g = model.median_g

    # expected median
    x = torch.from_numpy(adata.X)
    if transform is not None:
        x = transform(x)
    for i in range(x.shape[1]):
        mask = torch.nonzero(x[:, i])
        expected_median = torch.median(x[mask, i])
        actual_median = actual_median_g[i]
        # assert within 1% accuracy
        np.testing.assert_allclose(expected_median, actual_median, rtol=0.01)


def test_load_from_checkpoint_multi_device():
    # using pytest tmp_path fixture creates a directory per rank
    # so we need to use a fixed shared directory
    tmp_path = Path("/tmp/test_load_from_checkpoint_multi_device")
    n, g = 4, 3
    devices = int(os.environ.get("TEST_DEVICES", "1"))
    # dataloader
    train_loader = torch.utils.data.DataLoader(
        BoringDataset(np.arange(n * g).reshape(n, g)),
        collate_fn=collate_fn,
    )
    # model
    init_args = {"g_genes": g, "target_count": 10}
    model = TDigestFromCLI(**init_args)
    config = {
        "model": {
            "model": {
                "class_path": "cellarium.ml.models.TDigestFromCLI",
                "init_args": init_args,
            }
        }
    }
    module = CellariumModule(model, config=config)
    # trainer
    trainer = pl.Trainer(
        accelerator="cpu",
        devices=devices,
        max_epochs=1,
        default_root_dir=tmp_path,
    )
    # fit
    trainer.fit(module, train_dataloaders=train_loader)

    # run tests only for rank 0
    if trainer.global_rank != 0:
        return

    # load model from checkpoint
    ckpt_path = tmp_path / f"lightning_logs/version_0/checkpoints/epoch=0-step={math.ceil(n / devices)}.ckpt"
    assert ckpt_path.is_file()
    loaded_model: TDigestFromCLI = CellariumModule.load_from_checkpoint(ckpt_path).model
    # assert
    assert isinstance(model.transform, NormalizeTotal)
    assert isinstance(loaded_model.transform, NormalizeTotal)
    assert model.transform.target_count == loaded_model.transform.target_count
    np.testing.assert_allclose(model.median_g, loaded_model.median_g)

    shutil.rmtree(tmp_path)



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause


from collections.abc import Sequence

import lightning.pytorch as pl
import numpy as np
import pyro
import pyro.distributions as dist
import torch

from cellarium.ml.models.model import CellariumModel
from cellarium.ml.utilities.testing import (
    assert_arrays_equal,
    assert_columns_and_array_lengths_equal,
)


class LogisticRegression(CellariumModel):
    """
    Logistic regression model.

    Args:
        n_obs:
            Number of observations.
        feature_schema:
            The variable names schema for the input data validation.
        c_categories:
            Number of categories.
        W_prior_scale:
            The scale of the Laplace prior for the weights.
        W_init_scale:
            Initialization scale for the ``W_gc`` parameter.
        seed:
            Random seed used to initialize parameters.
        log_metrics:
            Whether to log the histogram of the ``W_gc`` parameter.
    """

    def __init__(
        self,
        n_obs: int,
        feature_schema: Sequence[str],
        c_categories: int,
        W_prior_scale: float = 1.0,
        W_init_scale: float = 1.0,
        seed: int = 0,
        log_metrics: bool = True,
    ) -> None:
        super().__init__()

        # data
        self.n_obs = n_obs
        self.feature_schema = np.array(feature_schema)
        self.g_features = len(feature_schema)
        self.c_categories = c_categories

        rng = torch.Generator()
        rng.manual_seed(seed)
        # parameters
        self.W_prior_scale: torch.Tensor
        self.register_buffer("W_prior_scale", torch.tensor(W_prior_scale))
        self.W_gc = torch.nn.Parameter(W_init_scale * torch.randn((self.g_features, c_categories), generator=rng))
        self.b_c = torch.nn.Parameter(torch.zeros(c_categories))

        # loss
        self.elbo = pyro.infer.Trace_ELBO()

        self.log_metrics = log_metrics

    @staticmethod
    def _get_fn_args_from_batch(tensor_dict: dict[str, np.ndarray | torch.Tensor]) -> tuple[tuple, dict]:
        x_ng = tensor_dict["x_ng"]
        feature_g = tensor_dict["feature_g"]
        y_n = tensor_dict["y_n"]
        return (x_ng, feature_g, y_n), {}

    def forward(self, x_ng: torch.Tensor, feature_g: np.ndarray, y_n: torch.Tensor) -> torch.Tensor:
        """
        Args:
            x_ng:
                The input data.
            feature_g:
                The variable names for the input data.
            y_n:
                The target data.

        Returns:
            The loss.
        """
        assert_columns_and_array_lengths_equal("x_ng", x_ng, "feature_g", feature_g)
        assert_arrays_equal("feature_g", feature_g, "feature_schema", self.feature_schema)
        return self.elbo.differentiable_loss(self.model, self.guide, x_ng, y_n)

    def model(self, x_ng: torch.Tensor, y_n: torch.Tensor) -> None:
        W_gc = pyro.sample(
            "W",
            dist.Laplace(0, self.W_prior_scale).expand([self.g_features, self.c_categories]).to_event(2),
        )
        with pyro.plate("batch", size=self.n_obs, subsample_size=x_ng.shape[0]):
            logits_nc = x_ng @ W_gc + self.b_c
            pyro.sample("y", dist.Categorical(logits=logits_nc), obs=y_n)

    def guide(self, x_ng: torch.Tensor, y_n: torch.Tensor) -> None:
        pyro.sample("W", dist.Delta(self.W_gc).to_event(2))

    def on_batch_end(self, trainer: pl.Trainer) -> None:
        if trainer.global_rank != 0:
            return

        if not self.log_metrics:
            return

        if (trainer.global_step + 1) % trainer.log_every_n_steps != 0:  # type: ignore[attr-defined]
            return

        for logger in trainer.loggers:
            if isinstance(logger, pl.loggers.TensorBoardLogger):
                logger.experiment.add_histogram(
                    "W_gc",
                    self.W_gc,
                    global_step=trainer.global_step,
                )



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

import re
import shutil
import tempfile
import urllib.request

from anndata import AnnData, read_h5ad
from google.cloud.storage import Client

url_schemes = ("http:", "https:", "ftp:")


def read_h5ad_gcs(filename: str, storage_client: Client | None = None) -> AnnData:
    r"""
    Read ``.h5ad``-formatted hdf5 file from the Google Cloud Storage.

    Example::

        >>> adata = read_h5ad_gcs("gs://dsp-cellarium-cas-public/test-data/test_0.h5ad")

    Args:
        filename: Path to the data file in Cloud Storage.
    """
    assert filename.startswith("gs:")
    # parse bucket and blob names from the filename
    filename = re.sub(r"^gs://?", "", filename)
    bucket_name, blob_name = filename.split("/", 1)

    if storage_client is None:
        storage_client = Client()

    bucket = storage_client.bucket(bucket_name)
    blob = bucket.blob(blob_name)

    with blob.open("rb") as f:
        return read_h5ad(f)


def read_h5ad_url(filename: str) -> AnnData:
    r"""
    Read ``.h5ad``-formatted hdf5 file from the URL.

    Example::

        >>> adata = read_h5ad_url(
        ...     "https://storage.googleapis.com/dsp-cellarium-cas-public/test-data/test_0.h5ad"
        ... )

    Args:
        filename: URL of the data file.
    """
    assert any(filename.startswith(scheme) for scheme in url_schemes)
    with urllib.request.urlopen(filename) as response:
        with tempfile.TemporaryFile() as tmp_file:
            shutil.copyfileobj(response, tmp_file)
            return read_h5ad(tmp_file)


def read_h5ad_local(filename: str) -> AnnData:
    r"""
    Read ``.h5ad``-formatted hdf5 file from the local disk.

    Args:
        filename: Path to the local data file.
    """
    assert filename.startswith("file:")
    filename = re.sub(r"^file://?", "", filename)
    return read_h5ad(filename)


def read_h5ad_file(filename: str, **kwargs) -> AnnData:
    r"""
    Read ``.h5ad``-formatted hdf5 file from a filename.

    Args:
        filename: Path to the data file.
    """
    if filename.startswith("gs:"):
        return read_h5ad_gcs(filename, **kwargs)

    if filename.startswith("file:"):
        return read_h5ad_local(filename)

    if any(filename.startswith(scheme) for scheme in url_schemes):
        return read_h5ad_url(filename)

    return read_h5ad(filename)



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

import numpy as np
import pytest
import torch

from cellarium.ml.transforms import Filter, Log1p, NormalizeTotal, ZScore

n, g, target_count = 100, 3, 10_000


@pytest.fixture
def x_ng():
    rng = torch.Generator()
    rng.manual_seed(1465)
    rates = torch.rand((n, g), generator=rng) * 50
    x_ng = torch.poisson(rates.float(), generator=rng)
    return x_ng


@pytest.fixture
def log_normalize(x_ng: torch.Tensor):
    l_n1 = x_ng.sum(dim=-1, keepdim=True)
    y_ng = torch.log1p(target_count * x_ng / l_n1)
    mean_g = y_ng.mean(dim=0)
    std_g = y_ng.std(dim=0)
    feature_schema = [f"gene_{i}" for i in range(g)]
    transform = torch.nn.Sequential(
        NormalizeTotal(target_count),
        Log1p(),
        ZScore(mean_g, std_g, feature_schema),
    )
    return transform


def test_log_normalize_shape(x_ng: torch.Tensor, log_normalize: torch.nn.Sequential):
    new_x_ng = log_normalize(x_ng)
    assert x_ng.shape == new_x_ng.shape


def test_log_normalize_mean_std(x_ng: torch.Tensor, log_normalize: torch.nn.Sequential):
    new_x_ng = log_normalize(x_ng)

    actual_mean = new_x_ng.mean(dim=0)
    actual_std = new_x_ng.std(dim=0)

    np.testing.assert_allclose(0, actual_mean, atol=1e-5)
    np.testing.assert_allclose(1, actual_std, atol=1e-5)


@pytest.mark.parametrize(
    "filter_list",
    [["gene_0"], ["gene_0", "gene_1"], ["gene_0", "gene_2"]],
)
def test_filter(x_ng: torch.Tensor, filter_list: list[str]):
    transform = Filter(filter_list)
    feature_g = np.array([f"gene_{i}" for i in range(g)])
    new_x_ng = transform(x_ng, feature_g)
    assert new_x_ng.shape[1] == len(filter_list)
    assert new_x_ng.shape[0] == x_ng.shape[0]


def test_filter_cache():
    filter_list = ["gene_0", "gene_1"]
    transform = Filter(filter_list)
    transform.filter.cache_clear()

    m = 4
    for g in range(1, 1 + m):
        feature_g = np.array([f"gene_{i}" for i in range(g)])
        x_ng = torch.zeros((2, g))
        for _ in range(g):
            transform(x_ng, feature_g)

    assert transform.filter.cache_info().currsize == m
    assert transform.filter.cache_info().misses == m
    assert transform.filter.cache_info().hits == m * (m - 1) / 2



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

from typing import Any

import numpy as np
import pyro
import pyro.distributions as dist
import torch
from pyro.nn import PyroParam
from torch.distributions import constraints

from cellarium.ml.models.model import CellariumModel


class ProbabilisticPCA(CellariumModel):
    """
    Probabilistic PCA implemented in Pyro.

    Two flavors of probabilistic PCA are available - marginalized pPCA [1]
    and linear VAE [2].

    **References:**

    1. `Probabilistic Principal Component Analysis (Tipping et al.)
       <https://www.robots.ox.ac.uk/~cvrg/hilary2006/ppca.pdf>`_.
    2. `Understanding Posterior Collapse in Generative Latent Variable Models (Lucas et al.)
       <https://openreview.net/pdf?id=r1xaVLUYuE>`_.

    Args:
        n_cells:
            Number of cells.
        g_genes:
            Number of genes.
        k_components:
            Number of principal components.
        ppca_flavor:
            Type of the PPCA model. Has to be one of `marginalized` or `linear_vae`.
        mean_g:
            Mean gene expression of the input data. If ``None`` then the mean is
            set to a learnable parameter.
        W_init_scale:
            Scale of the random initialization of the `W_kg` parameter.
        sigma_init_scale:
            Initialization value of the `sigma` parameter.
        seed:
            Random seed used to initialize parameters.
        transform:
            If not ``None`` is used to transform the input data.
        elbo:
            ELBO loss function. Should be a subclass of :class:`~pyro.infer.ELBO`.
            If ``None``, defaults to :class:`~pyro.infer.Trace_ELBO`.
    """

    def __init__(
        self,
        n_cells: int,
        g_genes: int,
        k_components: int,
        ppca_flavor: str,
        mean_g: float | torch.Tensor | None = None,
        W_init_scale: float = 1.0,
        sigma_init_scale: float = 1.0,
        seed: int = 0,
        transform: torch.nn.Module | None = None,
        elbo: pyro.infer.ELBO | None = None,
    ):
        super().__init__()

        self.n_cells = n_cells
        self.g_genes = g_genes
        self.k_components = k_components
        assert ppca_flavor in [
            "marginalized",
            "linear_vae",
        ], "ppca_flavor must be one of 'marginalized' or 'linear_vae'"
        self.ppca_flavor = ppca_flavor
        self.transform = transform
        self.elbo = elbo or pyro.infer.Trace_ELBO()

        if isinstance(mean_g, torch.Tensor) and mean_g.dim():
            assert mean_g.shape == (
                g_genes,
            ), f"Expected meang_g to have a shape ({g_genes},) but found {mean_g.shape}."
        if mean_g is None:
            # make mean_g a learnable parameter
            self.mean_g = PyroParam(lambda: torch.zeros(g_genes))
        else:
            self.register_buffer("mean_g", torch.as_tensor(mean_g))

        rng = torch.Generator()
        rng.manual_seed(seed)
        # model parameters
        self.W_kg = PyroParam(lambda: W_init_scale * torch.randn((k_components, g_genes), generator=rng))
        self.sigma = PyroParam(lambda: torch.tensor(sigma_init_scale), constraint=constraints.positive)

    @staticmethod
    def _get_fn_args_from_batch(tensor_dict: dict[str, np.ndarray | torch.Tensor]) -> tuple[tuple, dict]:
        x = tensor_dict["X"]
        return (x,), {}

    def forward(self, *args: Any, **kwargs: Any) -> torch.Tensor:
        return self.elbo.differentiable_loss(self.model, self.guide, *args, **kwargs)

    def model(self, x_ng: torch.Tensor) -> None:
        if self.transform is not None:
            x_ng = self.transform(x_ng)

        with pyro.plate("cells", size=self.n_cells, subsample_size=x_ng.shape[0]):
            if self.ppca_flavor == "marginalized":
                pyro.sample(
                    "counts",
                    dist.LowRankMultivariateNormal(
                        loc=self.mean_g,
                        cov_factor=self.W_kg.T,
                        cov_diag=self.sigma**2 * x_ng.new_ones(self.g_genes),
                    ),
                    obs=x_ng,
                )
            else:
                z_nk = pyro.sample(
                    "z",
                    dist.Normal(x_ng.new_zeros(self.k_components), 1).to_event(1),
                )
                pyro.sample(
                    "counts",
                    dist.Normal(self.mean_g + z_nk @ self.W_kg, self.sigma).to_event(1),
                    obs=x_ng,
                )

    def guide(self, x_ng: torch.Tensor) -> None:
        if self.ppca_flavor == "marginalized":
            return

        if self.transform is not None:
            x_ng = self.transform(x_ng)

        with pyro.plate("cells", size=self.n_cells, subsample_size=x_ng.shape[0]):
            V_gk = torch.linalg.solve(self.M_kk, self.W_kg).T
            D_k = self.sigma / torch.sqrt(torch.diag(self.M_kk))
            pyro.sample("z", dist.Normal((x_ng - self.mean_g) @ V_gk, D_k).to_event(1))

    @torch.inference_mode()
    def get_latent_representation(
        self,
        x_ng: torch.Tensor,
    ) -> torch.Tensor:
        r"""
        Return the latent representation for each cell.

        .. note::
           Gradients are disabled, used for inference only.
        """
        V_gk = torch.linalg.solve(self.M_kk, self.W_kg).T
        return (x_ng - self.mean_g) @ V_gk

    @property
    def M_kk(self) -> torch.Tensor:
        return self.W_kg @ self.W_kg.T + self.sigma**2 * torch.eye(self.k_components, device=self.sigma.device)

    @property
    @torch.inference_mode()
    def L_k(self) -> torch.Tensor:
        r"""
        Vector with elements given by the PC eigenvalues.

        .. note::
           Gradients are disabled, used for inference only.
        """
        S_k = torch.linalg.svdvals(self.W_kg.T)
        return S_k**2 + self.sigma**2

    @property
    @torch.inference_mode()
    def U_gk(self) -> torch.Tensor:
        r"""
        Principal components corresponding to eigenvalues ``L_k``.

        .. note::
           Gradients are disabled, used for inference only.
        """
        return torch.linalg.svd(self.W_kg.T, full_matrices=False).U

    @property
    @torch.inference_mode()
    def W_variance(self) -> float:
        r"""
        .. note::
           Gradients are disabled, used for inference only.
        """
        return torch.trace(self.W_kg.T @ self.W_kg).item()

    @property
    @torch.inference_mode()
    def sigma_variance(self) -> float:
        r"""
        .. note::
           Gradients are disabled, used for inference only.
        """
        return (self.g_genes * self.sigma**2).item()



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause


import numpy as np
import torch
from numpy.typing import ArrayLike
from torch import nn

from cellarium.ml.utilities.testing import (
    assert_arrays_equal,
    assert_columns_and_array_lengths_equal,
    assert_nonnegative,
)


class DivideByScale(nn.Module):
    """
    Divide gene counts by a scale.

    .. math::

        y_{ng} = \\frac{x_{ng}}{\\mathrm{scale}_g + \\mathrm{eps}}

    Args:
        scale_g:
            A scale for each gene.
        feature_schema:
            The variable names schema for the input data validation.
        eps:
            A value added to the denominator for numerical stability.
    """

    def __init__(self, scale_g: torch.Tensor, feature_schema: ArrayLike, eps: float = 1e-6) -> None:
        super().__init__()
        self.scale_g: torch.Tensor
        self.register_buffer("scale_g", scale_g)
        self.feature_schema = np.array(feature_schema)
        assert_nonnegative("eps", eps)
        self.eps = eps

    def forward(
        self,
        x_ng: torch.Tensor,
        feature_g: np.ndarray | None = None,
    ) -> torch.Tensor:
        """
        Args:
            x_ng:
                Gene counts.
            feature_g:
                The list of the variable names in the input data. If ``None``, no validation is performed.

        Returns:
            Gene counts divided by scale.
        """
        if feature_g is not None:
            assert_columns_and_array_lengths_equal("x_ng", x_ng, "feature_g", feature_g)
            assert_arrays_equal("feature_g", feature_g, "feature_schema", self.feature_schema)

        return x_ng / (self.scale_g + self.eps)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(scale_g={self.scale_g}, feature_schema={self.feature_schema}, eps={self.eps})"
        )



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

from cellarium.ml.callbacks.prediction_writer import PredictionWriter
from cellarium.ml.callbacks.variance_monitor import VarianceMonitor

__all__ = ["PredictionWriter", "VarianceMonitor"]



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

from cellarium.ml.data.dadc_dataset import IterableDistributedAnnDataCollectionDataset
from cellarium.ml.data.distributed_anndata import (
    DistributedAnnDataCollection,
    DistributedAnnDataCollectionView,
    LazyAnnData,
)
from cellarium.ml.data.fileio import read_h5ad_file, read_h5ad_gcs, read_h5ad_local, read_h5ad_url
from cellarium.ml.data.schema import AnnDataSchema

__all__ = [
    "AnnDataSchema",
    "DistributedAnnDataCollection",
    "DistributedAnnDataCollectionView",
    "IterableDistributedAnnDataCollectionDataset",
    "LazyAnnData",
    "read_h5ad_file",
    "read_h5ad_gcs",
    "read_h5ad_local",
    "read_h5ad_url",
]



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

import os
from typing import Any

import pytest

from cellarium.ml.cli import main

devices = os.environ.get("TEST_DEVICES", "1")

CONFIGS = [
    {
        "model_name": "geneformer",
        "subcommand": "fit",
        "fit": {
            "model": {
                "model": {
                    "class_path": "cellarium.ml.models.GeneformerFromCLI",
                    "init_args": {
                        "hidden_size": "2",
                        "num_hidden_layers": "1",
                        "num_attention_heads": "1",
                        "intermediate_size": "4",
                        "max_position_embeddings": "2",
                    },
                },
            },
            "data": {
                "filenames": "https://storage.googleapis.com/dsp-cellarium-cas-public/test-data/test_0.h5ad",
                "shard_size": "100",
                "max_cache_size": "2",
                "batch_keys": {
                    "X": {
                        "attr": "X",
                        "convert_fn": "cellarium.ml.utilities.data.densify",
                    },
                    "var_names": {"attr": "var_names"},
                },
                "batch_size": "5",
                "num_workers": "1",
            },
            "trainer": {
                "accelerator": "cpu",
                "devices": devices,
                "max_steps": "1",
            },
        },
    },
    {
        "model_name": "geneformer",
        "subcommand": "predict",
        "predict": {
            "model": {
                "model": {
                    "class_path": "cellarium.ml.models.GeneformerFromCLI",
                    "init_args": {
                        "hidden_size": "2",
                        "num_hidden_layers": "1",
                        "num_attention_heads": "1",
                        "intermediate_size": "4",
                        "max_position_embeddings": "2",
                    },
                },
            },
            "data": {
                "filenames": "https://storage.googleapis.com/dsp-cellarium-cas-public/test-data/test_0.h5ad",
                "shard_size": "100",
                "max_cache_size": "2",
                "batch_keys": {
                    "X": {
                        "attr": "X",
                        "convert_fn": "cellarium.ml.utilities.data.densify",
                    },
                    "var_names": {"attr": "var_names"},
                    "obs_names": {"attr": "obs_names"},
                },
                "batch_size": "5",
                "num_workers": "1",
            },
            "trainer": {
                "accelerator": "cpu",
                "devices": devices,
                "max_steps": "1",
                "limit_predict_batches": "1",
            },
            "return_predictions": "false",
        },
    },
    {
        "model_name": "probabilistic_pca",
        "subcommand": "fit",
        "fit": {
            "model": {
                "model": "cellarium.ml.models.ProbabilisticPCAFromCLI",
            },
            "data": {
                "filenames": "https://storage.googleapis.com/dsp-cellarium-cas-public/test-data/test_{0..1}.h5ad",
                "shard_size": "100",
                "max_cache_size": "2",
                "batch_keys": {
                    "X": {
                        "attr": "X",
                        "convert_fn": "cellarium.ml.utilities.data.densify",
                    },
                },
                "batch_size": "50",
                "shuffle": "true",
                "num_workers": "2",
            },
            "trainer": {
                "accelerator": "cpu",
                "devices": devices,
                "max_steps": "4",
            },
        },
    },
    {
        "model_name": "onepass_mean_var_std",
        "subcommand": "fit",
        "fit": {
            "model": {
                "model": "cellarium.ml.models.OnePassMeanVarStdFromCLI",
            },
            "data": {
                "filenames": "https://storage.googleapis.com/dsp-cellarium-cas-public/test-data/test_{0..1}.h5ad",
                "shard_size": "100",
                "max_cache_size": "2",
                "batch_keys": {
                    "X": {
                        "attr": "X",
                        "convert_fn": "cellarium.ml.utilities.data.densify",
                    },
                },
                "batch_size": "50",
                "num_workers": "2",
            },
            "trainer": {
                "accelerator": "cpu",
                "devices": devices,
            },
        },
    },
    {
        "model_name": "incremental_pca",
        "subcommand": "fit",
        "fit": {
            "model": {
                "model": {
                    "class_path": "cellarium.ml.models.IncrementalPCAFromCLI",
                    "init_args": {"k_components": "50"},
                },
            },
            "data": {
                "filenames": "https://storage.googleapis.com/dsp-cellarium-cas-public/test-data/test_{0..1}.h5ad",
                "shard_size": "100",
                "max_cache_size": "2",
                "batch_keys": {
                    "X": {
                        "attr": "X",
                        "convert_fn": "cellarium.ml.utilities.data.densify",
                    },
                },
                "batch_size": "50",
                "num_workers": "2",
            },
            "trainer": {
                "accelerator": "cpu",
                "devices": devices,
            },
        },
    },
    {
        "model_name": "incremental_pca",
        "subcommand": "predict",
        "predict": {
            "model": {
                "model": {
                    "class_path": "cellarium.ml.models.IncrementalPCAFromCLI",
                    "init_args": {"k_components": "50"},
                },
            },
            "data": {
                "filenames": "https://storage.googleapis.com/dsp-cellarium-cas-public/test-data/test_{0..1}.h5ad",
                "shard_size": "100",
                "max_cache_size": "2",
                "batch_keys": {
                    "X": {
                        "attr": "X",
                        "convert_fn": "cellarium.ml.utilities.data.densify",
                    },
                    "obs_names": {"attr": "obs_names"},
                },
                "batch_size": "50",
                "num_workers": "2",
            },
            "trainer": {
                "accelerator": "cpu",
                "devices": devices,
                "callbacks": {
                    "class_path": "cellarium.ml.callbacks.PredictionWriter",
                    "init_args": {"output_dir": "./output"},
                },
            },
            "return_predictions": "false",
        },
    },
    {
        "model_name": "logistic_regression",
        "subcommand": "fit",
        "fit": {
            "model": {
                "model": "cellarium.ml.models.LogisticRegression",
            },
            "data": {
                "filenames": "https://storage.googleapis.com/dsp-cellarium-cas-public/test-data/test_{0..1}.h5ad",
                "shard_size": "100",
                "max_cache_size": "2",
                "batch_keys": {
                    "x_ng": {
                        "attr": "X",
                        "convert_fn": "cellarium.ml.utilities.data.densify",
                    },
                    "feature_g": {
                        "attr": "var_names",
                    },
                    "y_n": {
                        "attr": "obs",
                        "key": "cell_type",
                        "convert_fn": "cellarium.ml.utilities.data.categories_to_codes",
                    },
                },
                "batch_size": "50",
                "shuffle": "true",
                "num_workers": "2",
            },
            "trainer": {
                "accelerator": "cpu",
                "devices": devices,
                "max_steps": "4",
            },
        },
    },
    {
        "model_name": "tdigest",
        "subcommand": "fit",
        "fit": {
            "model": {"model": "cellarium.ml.models.TDigestFromCLI"},
            "data": {
                "filenames": "https://storage.googleapis.com/dsp-cellarium-cas-public/test-data/test_{0..1}.h5ad",
                "shard_size": "100",
                "max_cache_size": "2",
                "batch_keys": {
                    "X": {
                        "attr": "X",
                        "convert_fn": "cellarium.ml.utilities.data.densify",
                    },
                },
                "batch_size": "50",
                "num_workers": "2",
            },
            "trainer": {
                "accelerator": "cpu",
                "devices": devices,
            },
        },
    },
]


@pytest.mark.parametrize("config", CONFIGS)
def test_cpu_multi_device(config: dict[str, Any]):
    if config["subcommand"] == "predict":
        assert config["predict"]["return_predictions"] == "false"
    main(config)



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

import math
import os
from pathlib import Path

import lightning.pytorch as pl
import numpy as np
import torch

from cellarium.ml import CellariumModule
from cellarium.ml.models import GeneformerFromCLI
from cellarium.ml.utilities.data import collate_fn
from tests.common import BoringDataset


def test_load_from_checkpoint_multi_device(tmp_path: Path):
    n, g = 4, 3
    var_names = np.array([f"gene_{i}" for i in range(g)])
    devices = int(os.environ.get("TEST_DEVICES", "1"))
    # dataloader
    train_loader = torch.utils.data.DataLoader(
        BoringDataset(
            np.arange(n * g).reshape(n, g),
            var_names=var_names,
        ),
        collate_fn=collate_fn,
    )
    # model
    init_args = {
        "feature_schema": var_names,
        "hidden_size": 2,
        "num_hidden_layers": 1,
        "num_attention_heads": 1,
        "intermediate_size": 4,
        "max_position_embeddings": 2,
    }
    model = GeneformerFromCLI(**init_args)  # type: ignore[arg-type]
    config = {
        "model": {
            "model": {
                "class_path": "cellarium.ml.models.GeneformerFromCLI",
                "init_args": init_args,
            }
        }
    }
    module = CellariumModule(model, config=config)
    # trainer
    trainer = pl.Trainer(
        accelerator="cpu",
        devices=devices,
        max_epochs=1,
        default_root_dir=tmp_path,
    )
    # fit
    trainer.fit(module, train_dataloaders=train_loader)

    # run tests only for rank 0
    if trainer.global_rank != 0:
        return

    # load model from checkpoint
    ckpt_path = tmp_path / f"lightning_logs/version_0/checkpoints/epoch=0-step={math.ceil(n / devices)}.ckpt"
    assert ckpt_path.is_file()
    loaded_model: GeneformerFromCLI = CellariumModule.load_from_checkpoint(ckpt_path).model
    # assert
    assert np.array_equal(model.feature_schema, loaded_model.feature_schema)
    np.testing.assert_allclose(model.feature_ids, loaded_model.feature_ids)



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

import math

import numpy as np
import torch
from torch.utils.data import IterableDataset

from cellarium.ml.data.distributed_anndata import DistributedAnnDataCollection
from cellarium.ml.utilities.data import AnnDataField, get_rank_and_num_replicas, get_worker_info


class IterableDistributedAnnDataCollectionDataset(IterableDataset):
    r"""
    Iterable DistributedAnnDataCollection Dataset.

    When :attr:`shuffle` is set to ``True`` then the iterator yields datapoints that are
    uniformly sampled from the entire dataset. Typical use cases include training variational
    models using the stochastic gradient descent algorithm.

    In order to maximize buffer usage, we only shuffle shards and datapoints within individual
    shards (and not across shards). Therefore, to achieve unbiased pseudo-random uniform sampling,
    it is imperative that the shards themselves contain datapoints that are uniformly sampled
    from the entire dataset. If correlations exist between datapoints in a given shard (e.g. all
    cells coming from the same tissue or experiment), then this assumption is violated. It is
    the user's responsibility to prepare appropriately shuffled data shards.

    Example::

        >>> from cellarium.ml.data import (
        ...     DistributedAnnDataCollection,
        ...     IterableDistributedAnnDataCollectionDataset,
        ... )
        >>> from cellarium.ml.utilities.data import AnnDataField, densify

        >>> dadc = DistributedAnnDataCollection(
        ...     "gs://bucket-name/folder/adata{000..005}.h5ad",
        ...     shard_size=10_000,
        ...     max_cache_size=2)

        >>> dataset = IterableDistributedAnnDataCollectionDataset(
        ...     dadc,
        ...     batch_keys={
        ...         "x_ng": AnnDataField(attr="X", convert_fn=densify),
        ...         "feature_g": AnnDataField(attr="var_names"),
        ...     },
        ...     batch_size=5000,
        ...     shuffle=True,
        ...     seed=0,
        ...     drop_last=True,
        ... )

    Args:
        dadc:
            DistributedAnnDataCollection from which to load the data.
        batch_keys:
            Dictionary that specifies which attributes and keys of the :attr:`dadc` to return
            in the batch data and how to convert them. Keys must correspond to
            the input keys of the transforms or the model. Values must be instances of
            :class:`cellarium.ml.utilities.data.AnnDataField`.
        batch_size:
            How many samples per batch to load.
        shuffle:
            If ``True``, the data is reshuffled at every epoch.
        seed:
            Random seed used to shuffle the sampler if :attr:`shuffle=True`.
        drop_last:
            If ``True``, then the sampler will drop the tail of the data
            to make it evenly divisible across the number of replicas. If ``False``,
            the sampler will add extra indices to make the data evenly divisible across
            the replicas.
        test_mode:
            If ``True``, then tracking of cache and worker informations will be enabled.
    """

    def __init__(
        self,
        dadc: DistributedAnnDataCollection,
        batch_keys: dict[str, AnnDataField],
        batch_size: int = 1,
        shuffle: bool = False,
        seed: int = 0,
        drop_last: bool = False,
        test_mode: bool = False,
    ) -> None:
        self.dadc = dadc
        self.batch_keys = batch_keys
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.seed = seed
        self.drop_last = drop_last
        self.epoch = 0
        self.test_mode = test_mode

    def __len__(self) -> int:
        """
        Returns the number of batches per replica.
        """
        _, num_replicas = get_rank_and_num_replicas()

        if self.drop_last and len(self.dadc) % num_replicas != 0:
            # Split to nearest available length that is evenly divisible.
            # This is to ensure each rank receives the same amount of data.
            per_replica = len(self.dadc) // num_replicas
        else:
            per_replica = math.ceil(len(self.dadc) / num_replicas)
        return math.ceil(per_replica / float(self.batch_size))

    def set_epoch(self, epoch: int) -> None:
        r"""
        Sets the epoch for the iterator. When :attr:`shuffle=True`, this ensures all replicas
        use a different random ordering for each epoch.
        """
        self.epoch = epoch

    def __getitem__(self, idx: int | list[int] | slice) -> dict[str, np.ndarray]:
        r"""
        Returns a dictionary containing the data from the :attr:`dadc` with keys specified by its :attr:`dadc.convert`
        at the given index ``idx``.
        """

        data = {}
        for key, field in self.batch_keys.items():
            data[key] = field(self.dadc)[idx]

        # for testing purposes
        if self.test_mode:
            rank, num_replicas = get_rank_and_num_replicas()
            worker_id, num_workers = get_worker_info()
            data["rank"] = np.array([rank])
            data["num_replicas"] = np.array([num_replicas])
            data["worker_id"] = np.array([worker_id])
            data["num_workers"] = np.array([num_workers])
            data["miss_count"] = np.array([self.dadc.cache.miss_count])
            data["epoch"] = np.array([self.epoch])

        return data

    def __iter__(self):
        r"""
        Iterate through the dataset by trying to minimize the amount of anndata files
        fetched by each worker.

        .. note::
            Returned iterator is determined by the ``torch.utils.data.get_worker_info()``
            and ``torch.distributed`` contexts. Iterated indices are evenly divided between replicas
            (see :attr:`drop_last`). If multiple workers per replica, then indices are further
            divided between workers (last worker might contain less indices than other workers, see
            examples below). Indices are shuffled and iterated in a manner that minimizes the overlap
            between the data chunks loaded by each worker.

        Example 1::

            indices=[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]
            n_obs=12
            num_replicas=1
            batch_size=2
            num_workers=3
            num_batches=6
            batches_per_worker=2
            per_worker=4

        +----------+-------+---------+
        |          |batch 0| batch 1 |
        +==========+=======+=========+
        | worker 0 | (0,1) | (2,3)   |
        +----------+-------+---------+
        | worker 1 | (4,5) | (6,7)   |
        +----------+-------+---------+
        | worker 2 | (8,9) | (10,11) |
        +----------+-------+---------+


        Example 2::

            indices=[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
            n_obs=11
            num_replicas=1
            batch_size=2
            num_workers=2
            num_batches=6
            batches_per_worker=3
            per_worker=6

        +----------+-------+-------+-------+
        |          |batch 0|batch 1|batch 2|
        +==========+=======+=======+=======+
        | worker 0 | (0,1) | (2,3) | (4,5) |
        +----------+-------+-------+-------+
        | worker 1 | (6,7) | (8,9) | (10,) |
        +----------+-------+-------+-------+


        Example 3::

            indices=[0, 1, 2, 3, 4, 5, 6, 7]
            n_obs=8
            num_replicas=1
            batch_size=3
            num_workers=2
            num_batches=3
            batches_per_worker=2
            per_worker=6

        +----------+---------+---------+
        |          | batch 0 | batch 1 |
        +==========+=========+=========+
        | worker 0 | (0,1,2) | (3,4,5) |
        +----------+---------+---------+
        | worker 1 | (6,7)   |         |
        +----------+---------+---------+


        Example 4::

            indices=[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
            n_obs=11
            num_replicas=2
            drop_last=True

            truncated_indices=[0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
            total_size=10

            first_replica=[0, 1, 2, 3, 4]
            batch_size=2
            num_workers=1
            num_batches=3
            batches_per_worker=3
            per_worker=6

            second_replica=[5, 6, 7, 8, 9]
            batch_size=2
            num_workers=1
            num_batches=3
            batches_per_worker=3
            per_worker=6

        *Replica 1*

        +----------+-------+-------+-------+
        |          |batch 0|batch 1|batch 2|
        +==========+=======+=======+=======+
        | worker 0 | (0,1) | (2,3) | (4,)  |
        +----------+-------+-------+-------+

        *Replica 2*

        +----------+-------+-------+-------+
        |          |batch 0|batch 1|batch 2|
        +==========+=======+=======+=======+
        | worker 0 | (5,6) | (7,8) | (9,)  |
        +----------+-------+-------+-------+


        Example 5::

            indices=[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
            n_obs=11
            num_replicas=2
            drop_last=False

            padded_indices=[0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 0]
            total_size=12

            first_replica=[0, 1, 2, 3, 4, 5]
            batch_size=2
            num_workers=1
            num_batches=3
            batches_per_worker=3
            per_worker=6

            second_replica=[6, 7, 8, 9, 10, 0]
            batch_size=2
            num_workers=1
            num_batches=3
            batches_per_worker=3
            per_worker=6

        *Replica 1*

        +----------+-------+-------+-------+
        |          |batch 0|batch 1|batch 2|
        +==========+=======+=======+=======+
        | worker 0 | (0,1) | (2,3) | (4,5) |
        +----------+-------+-------+-------+

        *Replica 2*

        +----------+-------+-------+--------+
        |          |batch 0|batch 1|batch 2 |
        +==========+=======+=======+========+
        | worker 0 | (6,7) | (8,9) | (10,0) |
        +----------+-------+-------+--------+
        """
        if self.test_mode:
            # clear lru cache
            self.dadc.cache.clear()

        # replicas
        rank, num_replicas = get_rank_and_num_replicas()

        if self.drop_last and len(self.dadc) % num_replicas != 0:
            # Split to nearest available length that is evenly divisible.
            # This is to ensure each rank receives the same amount of data.
            per_replica = len(self.dadc) // num_replicas
        else:
            per_replica = math.ceil(len(self.dadc) / num_replicas)
        total_size = per_replica * num_replicas
        batches_per_replica = math.ceil(per_replica / float(self.batch_size))

        # workers
        worker_id, num_workers = get_worker_info()

        batches_per_worker = math.ceil(batches_per_replica / float(num_workers))
        per_worker = batches_per_worker * self.batch_size

        # split workload
        iter_start = worker_id * per_worker
        iter_end = min(iter_start + per_worker, per_replica)

        # indices
        if self.shuffle:
            rng = torch.Generator()
            rng.manual_seed(self.seed + self.epoch)
            iter_limits = list(zip([0] + self.dadc.limits, self.dadc.limits))
            # shuffle shards
            limit_indices = torch.randperm(len(iter_limits), generator=rng).tolist()
            indices = []
            for limit_idx in limit_indices:
                lower, upper = iter_limits[limit_idx]
                # shuffle cells within shards
                indices.extend((torch.randperm(upper - lower, generator=rng) + lower).tolist())
        else:
            indices = list(range(len(self.dadc)))

        if not self.drop_last:
            # add extra samples to make it evenly divisible
            padding_size = total_size - len(indices)
            if padding_size <= len(indices):
                indices += indices[:padding_size]
            else:
                indices += (indices * math.ceil(padding_size / len(indices)))[:padding_size]
        else:
            # remove tail of data to make it evenly divisible.
            indices = indices[:total_size]
        indices = indices[rank * per_replica : (rank + 1) * per_replica]
        assert len(indices) == per_replica

        yield from (self[indices[i : i + self.batch_size]] for i in range(iter_start, iter_end, self.batch_size))
        # Sets epoch for persistent workers
        self.set_epoch(self.epoch + 1)



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

from collections.abc import Sequence

import numpy as np
from anndata import AnnData


class AnnDataSchema:
    r"""
    Store reference AnnData attributes for a collection of distributed AnnData objects.

    Validate AnnData objects against reference attributes.

    Example::

        >>> ref_adata = AnnData(X=np.zeros(2, 3))
        >>> ref_adata.obs["batch"] = [0, 1]
        >>> ref_adata.var["mu"] = ["a", "b", "c"]
        >>> adata = AnnData(X=np.zeros(4, 3))
        >>> adata.obs["batch"] = [2, 3, 4, 5]
        >>> adata.var["mu"] = ["a", "b", "c"]
        >>> schema = AnnDataSchema(ref_adata)
        >>> schema.validate_anndata(adata)

    Args:
        adata:
            Reference AnnData object.
        obs_columns:
            Subset of columns to validate in the ``.obs`` attribute.
            If ``None``, all columns are validated.
    """

    attrs = ["obs", "obsm", "var", "varm", "varp", "var_names", "layers"]

    def __init__(self, adata: AnnData, obs_columns: Sequence | None = None) -> None:
        self.attr_values = {}
        for attr in self.attrs:
            # FIXME: some of the attributes have a reference to the anndata object itself.
            # This results in anndata object not being garbage collected.
            self.attr_values[attr] = getattr(adata, attr)
        self.obs_columns = obs_columns

    def validate_anndata(self, adata: AnnData) -> None:
        """Validate anndata has proper attributes."""

        for attr in self.attrs:
            value = getattr(adata, attr)
            ref_value = self.attr_values[attr]
            if attr == "obs":
                if self.obs_columns is not None:
                    # Subset the columns to validate
                    ref_value = ref_value[self.obs_columns]
                    value = value[self.obs_columns]
                # compare the elements inside the Index object and their order
                if not ref_value.columns.equals(value.columns):
                    raise ValueError(
                        ".obs attribute columns for anndata passed in does not match .obs attribute columns "
                        "of the reference anndata."
                    )
                if not ref_value.dtypes.equals(value.dtypes):
                    raise ValueError(
                        ".obs attribute dtypes for anndata passed in does not match .obs attribute dtypes "
                        "of the reference anndata."
                    )
            elif attr in ["var", "var_names"]:
                # For var compare if two DataFrames have the same shape and elements
                # and the same row/column index.
                # For var_names compare the elements inside the Index object and their order
                if not ref_value.equals(value):
                    raise ValueError(
                        f".{attr} attribute for anndata passed in does not match .{attr} attribute "
                        "of the reference anndata."
                    )
            elif attr in ["layers", "obsm", "varm", "varp"]:
                # compare the keys
                if not set(ref_value.keys()) == set(value.keys()):
                    raise ValueError(
                        f".{attr} attribute keys for anndata passed in does not match .{attr} attribute keys "
                        "of the reference anndata."
                    )
                if attr in ["varm", "varp"]:
                    for key in ref_value:
                        arr = value[key]
                        ref_arr = ref_value[key]
                        if not np.array_equal(ref_arr, arr):
                            raise ValueError(
                                f".{attr} attribute for anndata passed in does not match .{attr} attribute "
                                "of the reference anndata."
                            )



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

"""
Command line interface for Cellarium ML.
"""

import sys
import warnings
from collections.abc import Callable
from typing import Any

from jsonargparse import Namespace
from lightning.pytorch.cli import ArgsType, LightningArgumentParser, LightningCLI

from cellarium.ml import CellariumAnnDataDataModule, CellariumModule

REGISTERED_MODELS = {}


def register_model(model: Callable[[ArgsType], None]):
    REGISTERED_MODELS[model.__name__] = model
    return model


def lightning_cli_factory(
    model_class_path: str,
    link_arguments: list[tuple[str, str, Callable[[Any], object] | None]] | None = None,
    trainer_defaults: dict[str, Any] | None = None,
) -> type[LightningCLI]:
    """
    Factory function for creating a :class:`LightningCLI` with a preset model and custom argument linking.

    Example::

        cli = lightning_cli_factory(
            "cellarium.ml.models.IncrementalPCAFromCLI",
            link_arguments=[("data.n_vars", "model.model.init_args.g_genes", None)],
            trainer_defaults={
                "max_epochs": 1,  # one pass
                "strategy": {
                    "class_path": "lightning.pytorch.strategies.DDPStrategy",
                    "init_args": {"broadcast_buffers": False},
                },
            },
        )

    Args:
        model_class_path:
            A string representation of the model class path (e.g., ``"cellarium.ml.models.IncrementalPCAFromCLI"``).
        link_arguments:
            A list of tuples of the form ``(source, target)`` where ``source`` is linked to ``target``.
        trainer_defaults:
            Default values for the trainer.

    Returns:
        A :class:`LightningCLI` class with the given model and argument linking.
    """

    class NewLightningCLI(LightningCLI):
        def __init__(self, args: ArgsType = None) -> None:
            super().__init__(
                CellariumModule,
                CellariumAnnDataDataModule,
                trainer_defaults=trainer_defaults,
                args=args,
            )

        def instantiate_classes(self) -> None:
            super().instantiate_classes()

            # impute linked arguments after instantiation
            if link_arguments is not None:
                for source, target, compute_fn in link_arguments:
                    # e.g., source == "data.var_names"
                    source_key, *source_attrs = source.split(".")
                    # note that config_init is initialized, so value is an instance
                    config_init = self.config_init[self.subcommand]
                    value = config_init[source_key]
                    for attr in source_attrs:
                        value = getattr(value, attr)
                    if compute_fn is not None:
                        value = compute_fn(value)

                    # e.g., target == "model.model.init_args.feature_schema"
                    target_keys = target.split(".")
                    # note that config is dict-like, so assign the value to the last key
                    config = self.config[self.subcommand]
                    for key in target_keys[:-1]:
                        config = config[key]
                    config[target_keys[-1]] = value

            # save the config to the :attr:`model.hparams` to be able to load the model checkpoint
            self.model._set_hparams(self.config[self.subcommand])

        def add_arguments_to_parser(self, parser: LightningArgumentParser) -> None:
            if link_arguments is not None:
                for arg1, arg2, compute_fn in link_arguments:
                    parser.link_arguments(arg1, arg2, compute_fn=compute_fn, apply_on="instantiate")
            parser.set_defaults({"model.model": model_class_path})

    return NewLightningCLI


@register_model
def geneformer(args: ArgsType = None) -> None:
    r"""
    CLI to run the :class:`cellarium.ml.models.GeneformerFromCLI` model.

    This example shows how to fit feature count data to the Geneformer model [1].

    Example run::

        cellarium-ml geneformer fit \
            --data.filenames "gs://dsp-cellarium-cas-public/test-data/test_{0..3}.h5ad" \
            --data.shard_size 100 \
            --data.max_cache_size 2 \
            --data.batch_size 5 \
            --data.num_workers 1 \
            --trainer.accelerator gpu \
            --trainer.devices 1 \
            --trainer.default_root_dir runs/geneformer \
            --trainer.max_steps 10

    **References:**

    1. `Transfer learning enables predictions in network biology (Theodoris et al.)
       <https://www.nature.com/articles/s41586-023-06139-9>`_.

    Args:
        args: Arguments to parse. If ``None`` the arguments are taken from ``sys.argv``.
    """
    cli = lightning_cli_factory(
        "cellarium.ml.models.GeneformerFromCLI",
        link_arguments=[("data.var_names", "model.model.init_args.feature_schema", None)],
    )
    cli(args=args)


@register_model
def incremental_pca(args: ArgsType = None) -> None:
    r"""
    CLI to run the :class:`cellarium.ml.models.IncrementalPCAFromCLI` model.

    This example shows how to fit feature count data to incremental PCA
    model [1, 2].

    Example run::

        cellarium-ml incremental_pca fit \
            --model.model.init_args.k_components 50 \
            --data.filenames "gs://dsp-cellarium-cas-public/test-data/test_{0..3}.h5ad" \
            --data.shard_size 100 \
            --data.max_cache_size 2 \
            --data.batch_size 100 \
            --data.num_workers 4 \
            --trainer.accelerator gpu \
            --trainer.devices 1 \
            --trainer.default_root_dir runs/ipca \

    **References:**

    1. `A Distributed and Incremental SVD Algorithm for Agglomerative Data Analysis on Large Networks (Iwen et al.)
       <https://users.math.msu.edu/users/iwenmark/Papers/distrib_inc_svd.pdf>`_.
    2. `Incremental Learning for Robust Visual Tracking (Ross et al.)
       <https://www.cs.toronto.edu/~dross/ivt/RossLimLinYang_ijcv.pdf>`_.

    Args:
        args: Arguments to parse. If ``None`` the arguments are taken from ``sys.argv``.
    """
    cli = lightning_cli_factory(
        "cellarium.ml.models.IncrementalPCAFromCLI",
        link_arguments=[("data.n_vars", "model.model.init_args.g_genes", None)],
        trainer_defaults={
            "max_epochs": 1,  # one pass
            "strategy": {
                "class_path": "lightning.pytorch.strategies.DDPStrategy",
                "init_args": {"broadcast_buffers": False},
            },
        },
    )
    cli(args=args)


@register_model
def logistic_regression(args: ArgsType = None) -> None:
    r"""
    CLI to run the :class:`cellarium.ml.models.LogisticRegression` model.

    Example run::

        cellarium-ml logistic_regression fit \
            --data.filenames "gs://dsp-cellarium-cas-public/test-data/test_{0..3}.h5ad" \
            --data.shard_size 100 \
            --data.max_cache_size 2 \
            --data.batch_keys.x_ng.attr X \
            --data.batch_keys.x_ng.convert_fn cellarium.ml.utilities.data.densify \
            --data.batch_keys.feature_g.attr var_names \
            --data.batch_keys.y_n.attr obs \
            --data.batch_keys.y_n.key cell_type \
            --data.batch_keys.y_n.convert_fn cellarium.ml.utilities.data.categories_to_codes \
            --data.batch_size 100 \
            --data.num_workers 4 \
            --trainer.accelerator gpu \
            --trainer.devices 1 \
            --trainer.max_steps 1000

    Args:
        args: Arguments to parse. If ``None`` the arguments are taken from ``sys.argv``.
    """

    def get_c_categories(data: CellariumAnnDataDataModule) -> int:
        """
        Get the number of categories in the target variable.

        E.g. if the target variable is ``obs["cell_type"]`` then this function
        returns the number of categories in ``obs["cell_type"]``::

            >>> len(data.dadc[0].obs["cell_type"].cat.categories)
        """
        field = data.batch_keys["y_n"]
        value = getattr(data.dadc[0], field.attr)
        if field.key is not None:
            value = value[field.key]
        return len(value.cat.categories)

    cli = lightning_cli_factory(
        "cellarium.ml.models.LogisticRegression",
        link_arguments=[
            ("data.n_obs", "model.model.init_args.n_obs", None),
            ("data.var_names", "model.model.init_args.feature_schema", None),
            ("data", "model.model.init_args.c_categories", get_c_categories),
        ],
    )
    cli(args=args)


@register_model
def onepass_mean_var_std(args: ArgsType = None) -> None:
    r"""
    CLI to run the :class:`cellarium.ml.models.OnePassMeanVarStdFromCLI` model.

    This example shows how to calculate mean, variance, and standard deviation of log normalized
    feature count data in one pass [1].

    Example run::

        cellarium-ml onepass_mean_var_std fit \
            --data.filenames "gs://dsp-cellarium-cas-public/test-data/test_{0..3}.h5ad" \
            --data.shard_size 100 \
            --data.max_cache_size 2 \
            --data.batch_size 100 \
            --data.num_workers 4 \
            --trainer.accelerator gpu \
            --trainer.devices 1 \
            --trainer.default_root_dir runs/onepass \

    **References:**

    1. `Algorithms for calculating variance
       <https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance>`_.

    Args:
        args: Arguments to parse. If ``None`` the arguments are taken from ``sys.argv``.
    """
    cli = lightning_cli_factory(
        "cellarium.ml.models.OnePassMeanVarStdFromCLI",
        link_arguments=[("data.n_vars", "model.model.init_args.g_genes", None)],
        trainer_defaults={
            "max_epochs": 1,  # one pass
            "strategy": {
                "class_path": "lightning.pytorch.strategies.DDPStrategy",
                "init_args": {"broadcast_buffers": False},
            },
        },
    )
    cli(args=args)


@register_model
def probabilistic_pca(args: ArgsType = None) -> None:
    r"""
    CLI to run the :class:`cellarium.ml.models.ProbabilisticPCAFromCLI` model.

    This example shows how to fit feature count data to probabilistic PCA
    model [1].

    There are two flavors of probabilistic PCA model that are available:

    1. ``marginalized`` - latent variable ``z`` is marginalized out [1]. Marginalized
       model provides a closed-form solution for the marginal log-likelihood.
       Closed-form solution for the marginal log-likelihood has reduced
       variance compared to the ``linear_vae`` model.
    2. ``linear_vae`` - latent variable ``z`` has a diagonal Gaussian distribution [2].
       Training a linear VAE with variational inference recovers a uniquely identifiable
       global maximum  corresponding to the principal component directions.
       The global maximum of the ELBO objective for the linear VAE  is identical
       to the global maximum for the marginal log-likelihood of probabilistic PCA.

    Example run::

        cellarium-ml probabilistic_pca fit \
            --model.model.init_args.mean_var_std_ckpt_path \
            "runs/onepass/lightning_logs/version_0/checkpoints/module_checkpoint.pt" \
            --data.filenames "gs://dsp-cellarium-cas-public/test-data/test_{0..3}.h5ad" \
            --data.shard_size 100 \
            --data.max_cache_size 2 \
            --data.batch_size 100 \
            --data.shuffle true \
            --data.num_workers 4 \
            --trainer.accelerator gpu \
            --trainer.devices 1 \
            --trainer.max_steps 1000 \
            --trainer.default_root_dir runs/ppca \
            --trainer.callbacks cellarium.ml.callbacks.VarianceMonitor \
            --trainer.callbacks.mean_var_std_ckpt_path \
            "runs/onepass/lightning_logs/version_0/checkpoints/module_checkpoint.pt"

    **References:**

    1. `Probabilistic Principal Component Analysis (Tipping et al.)
       <https://www.robots.ox.ac.uk/~cvrg/hilary2006/ppca.pdf>`_.
    2. `Understanding Posterior Collapse in Generative Latent Variable Models (Lucas et al.)
       <https://openreview.net/pdf?id=r1xaVLUYuE>`_.

    Args:
        args: Arguments to parse. If ``None`` the arguments are taken from ``sys.argv``.
    """
    cli = lightning_cli_factory(
        "cellarium.ml.models.ProbabilisticPCAFromCLI",
        link_arguments=[
            ("data.n_obs", "model.model.init_args.n_cells", None),
            ("data.n_vars", "model.model.init_args.g_genes", None),
        ],
    )
    cli(args=args)


@register_model
def tdigest(args: ArgsType = None) -> None:
    r"""
    CLI to run the :class:`cellarium.ml.models.TDigestFromCLI` model.

    This example shows how to calculate non-zero median of normalized feature count
    data in one pass [1].

    Example run::

        cellarium-ml tdigest fit \
            --data.filenames "gs://dsp-cellarium-cas-public/test-data/test_{0..3}.h5ad" \
            --data.shard_size 100 \
            --data.max_cache_size 2 \
            --data.batch_size 100 \
            --data.num_workers 4 \
            --trainer.accelerator cpu \
            --trainer.devices 1 \
            --trainer.default_root_dir runs/tdigest \

    **References:**

    1. `Computing Extremely Accurate Quantiles Using T-Digests (Dunning et al.)
       <https://github.com/tdunning/t-digest/blob/master/docs/t-digest-paper/histo.pdf>`_.

    Args:
        args: Arguments to parse. If ``None`` the arguments are taken from ``sys.argv``.
    """
    cli = lightning_cli_factory(
        "cellarium.ml.models.TDigestFromCLI",
        link_arguments=[("data.n_vars", "model.model.init_args.g_genes", None)],
        trainer_defaults={
            "max_epochs": 1,  # one pass
        },
    )
    cli(args=args)


def main(args: ArgsType = None) -> None:
    """
    CLI that dispatches to the appropriate model cli based on the model name in ``args`` and runs it.

    Args:
        args: Arguments to parse. If ``None`` the arguments are taken from ``sys.argv``.
            The model name is expected to be the first argument if ``args`` is a list
            or the ``model_name`` key if ``args`` is a dictionary or ``Namespace``.
    """
    if isinstance(args, (dict, Namespace)):
        assert "model_name" in args, "'model_name' key must be specified in args"
        model_name = args.pop("model_name")
    elif isinstance(args, list):
        assert len(args) > 0, "'model_name' must be specified as the first argument in args"
        model_name = args.pop(0)
    elif args is None:
        args = sys.argv[1:].copy()
        assert len(args) > 0, "'model_name' must be specified after cellarium-ml"
        model_name = args.pop(0)

    assert (
        model_name in REGISTERED_MODELS
    ), f"'model_name' must be one of {list(REGISTERED_MODELS.keys())}. Got '{model_name}'"
    model_cli = REGISTERED_MODELS[model_name]
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message="LightningCLI's args parameter is intended to run from within Python")
        warnings.filterwarnings("ignore", message="Your `IterableDataset` has `__len__` defined.")
        model_cli(args)  # run the model


if __name__ == "__main__":
    main()



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

import os
import pickle
from collections.abc import Callable
from pathlib import Path
from typing import Any

import numpy as np
import pandas as pd
import pytest
from anndata import AnnData

from cellarium.ml.data import (
    DistributedAnnDataCollection,
    IterableDistributedAnnDataCollectionDataset,
    read_h5ad_file,
)
from cellarium.ml.utilities.data import AnnDataField


@pytest.fixture
def adatas_path(tmp_path: Path):
    n_cell, n_gene = (10, 5)
    # adata0.n_obs == 2, adata1.n_obs == 3, adata2.n_obs == 5
    limits = [2, 5, 10]

    rng = np.random.default_rng(1465)
    X = rng.integers(50, size=(n_cell, n_gene))
    L = rng.integers(50, size=(n_cell, n_gene))
    M = rng.normal(size=(n_cell, 2))
    V = rng.normal(size=(n_gene, 2))
    P = rng.normal(size=(n_gene, n_gene))
    obs = pd.DataFrame(
        {
            "A": rng.integers(0, 2, n_cell),
            "B": rng.integers(0, 2, n_cell),
        },
        index=[f"ref_cell{i:03d}" for i in range(n_cell)],
    )
    var = pd.DataFrame(
        {
            "D": np.zeros(n_gene),
            "E": np.ones(n_gene),
            "F": np.arange(n_gene),
        },
        index=[f"gene{i:03d}" for i in range(n_gene)],
    )
    adata = AnnData(
        X,
        dtype=X.dtype,
        obs=obs,
        var=var,
        layers={"L": L},
        obsm={"M": M},
        varm={"V": V},
        varp={"P": P},
    )
    adata.write(os.path.join(tmp_path, "adata.h5ad"))
    for i, limit in enumerate(zip([0] + limits, limits)):
        sliced_adata = adata[slice(*limit)]
        sliced_adata.write(os.path.join(tmp_path, f"adata.00{i}.h5ad"))
    return tmp_path


@pytest.fixture
def adt(adatas_path: Path):
    # single anndata
    adt = read_h5ad_file(str(os.path.join(adatas_path, "adata.h5ad")))
    return adt


@pytest.fixture(params=[(i, j) for i in (1, 2, 3) for j in (True, False)])
def dat(adatas_path: Path, request: pytest.FixtureRequest):
    # distributed anndata
    filenames = str(os.path.join(adatas_path, "adata.{000..002}.h5ad"))
    limits = [2, 5, 10]
    max_cache_size, cache_size_strictly_enforced = request.param
    dat = DistributedAnnDataCollection(
        filenames,
        limits,
        max_cache_size=max_cache_size,
        cache_size_strictly_enforced=cache_size_strictly_enforced,
    )
    return dat


def test_init_dat(dat: DistributedAnnDataCollection):
    # check that only one anndata was loaded during initialization
    assert len(dat.cache) == 1

    for i, ladata in enumerate(dat.adatas):
        # check that calling repr doesn't load anndata
        repr(ladata)
        if i == 0:
            assert ladata.cached is True
        else:
            assert ladata.cached is False

        # validate anndata
        adata = ladata.adata
        dat.schema.validate_anndata(adata)


@pytest.mark.parametrize("num_shards", [3, 4, 10])
@pytest.mark.parametrize("last_shard_size", [1, 2, 3, None])
def test_init_shard_size(adatas_path: Path, num_shards: int, last_shard_size: int | None):
    shard_size = 2
    filenames = str(os.path.join(adatas_path, f"adata.{{000..{num_shards-1:03}}}.h5ad"))
    dadc = DistributedAnnDataCollection(
        filenames,
        shard_size=shard_size,
        last_shard_size=last_shard_size,
        max_cache_size=1,
    )

    actual_len = len(dadc)
    expected_len = num_shards * shard_size
    if last_shard_size is not None:
        expected_len = expected_len + last_shard_size - shard_size

    assert actual_len == expected_len


@pytest.mark.parametrize(
    "row_select",
    [(slice(0, 2), 1), (slice(1, 4), 2), ([1, 2, 4, 4], 2), ([6, 1, 3], 3)],
    ids=["one adata", "two adatas", "sorted two adatas", "unsorted three adatas"],
)
@pytest.mark.parametrize("vidx", [slice(0, 2), [3, 1, 0]])
def test_indexing(
    adt: AnnData,
    dat: DistributedAnnDataCollection,
    row_select: tuple[slice | list, int],
    vidx: slice | list,
):
    # compare indexing single and distributed anndata
    max_cache_size = dat.max_cache_size
    cache_size_strictly_enforced = dat.cache_size_strictly_enforced
    oidx, n_adatas = row_select

    if cache_size_strictly_enforced and (n_adatas > max_cache_size):
        with pytest.raises(AssertionError, match="Expected the number of anndata files"):
            dat_view = dat[oidx, vidx]
    else:
        adt_view = adt[oidx, vidx]
        dat_view = dat[oidx, vidx]
        np.testing.assert_array_equal(adt_view.X, dat_view.X)
        np.testing.assert_array_equal(adt_view.var_names, dat_view.var_names)
        np.testing.assert_array_equal(adt_view.obs_names, dat_view.obs_names)
        np.testing.assert_array_equal(adt_view.layers["L"], dat_view.layers["L"])
        np.testing.assert_array_equal(adt_view.obsm["M"], dat_view.obsm["M"])
        np.testing.assert_array_equal(adt_view.obs["A"], dat_view.obs["A"])


def test_pickle(dat: DistributedAnnDataCollection):
    new_dat = pickle.loads(pickle.dumps(dat))

    assert len(new_dat.cache) == 0

    new_dat_view, dat_view = new_dat[:2], dat[:2]

    np.testing.assert_array_equal(new_dat_view.X, dat_view.X)
    np.testing.assert_array_equal(new_dat_view.var_names, dat_view.var_names)
    np.testing.assert_array_equal(new_dat_view.obs_names, dat_view.obs_names)
    np.testing.assert_array_equal(new_dat_view.layers["L"], dat_view.layers["L"])
    np.testing.assert_array_equal(new_dat_view.obsm["M"], dat_view.obsm["M"])
    np.testing.assert_array_equal(new_dat_view.obs["A"], dat_view.obs["A"])


@pytest.mark.parametrize(
    "row_select",
    [(slice(0, 2), 1), (slice(1, 4), 2), ([1, 2, 4, 4], 2), ([6, 1, 3], 3)],
    ids=["one adata", "two adatas", "sorted two adatas", "unsorted three adatas"],
)
def test_indexing_dataset(
    adt: AnnData,
    dat: DistributedAnnDataCollection,
    row_select: tuple[slice | list, int],
):
    # compare indexing single anndata and distributed anndata dataset
    max_cache_size = dat.max_cache_size
    cache_size_strictly_enforced = dat.cache_size_strictly_enforced
    oidx, n_adatas = row_select

    dataset = IterableDistributedAnnDataCollectionDataset(
        dat,
        batch_keys={
            "X": AnnDataField("X"),
            "obs_names": AnnDataField("obs_names"),
        },
    )

    if cache_size_strictly_enforced and (n_adatas > max_cache_size):
        with pytest.raises(AssertionError, match="Expected the number of anndata files"):
            dataset_X = dataset[oidx]["X"]
    else:
        adt_X = adt[oidx].X
        dataset_X = dataset[oidx]["X"]
        np.testing.assert_array_equal(adt_X, dataset_X)

        adt_obs_names = adt[oidx].obs_names
        dataset_obs_names = dataset[oidx]["obs_names"]
        np.testing.assert_array_equal(adt_obs_names, dataset_obs_names)


def test_pickle_dataset(dat: DistributedAnnDataCollection):
    dataset = IterableDistributedAnnDataCollectionDataset(
        dat,
        batch_keys={
            "X": AnnDataField("X"),
            "obs_names": AnnDataField("obs_names"),
        },
    )
    new_dataset = pickle.loads(pickle.dumps(dataset))

    assert len(new_dataset.dadc.cache) == 0

    np.testing.assert_array_equal(new_dataset[:2]["X"], dataset[:2]["X"])
    np.testing.assert_array_equal(new_dataset[:2]["obs_names"], dataset[:2]["obs_names"])


@pytest.fixture
def dadc(adatas_path: Path):
    # distributed anndata
    filenames = str(os.path.join(adatas_path, "adata.{000..002}.h5ad"))
    limits = [2, 5, 10]
    dadc = DistributedAnnDataCollection(
        filenames,
        limits,
        max_cache_size=2,
    )
    return dadc


@pytest.mark.parametrize(
    "attr,key,convert_fn",
    [
        ("X", None, None),
        ("obs", "A", None),
        ("obs_names", None, None),
        ("var_names", None, None),
        ("layers", "L", None),
    ],
)
@pytest.mark.parametrize("idx", [slice(0, 5), [0, 2, 4], 0])
def test_anndata_field(
    dadc: DistributedAnnDataCollection,
    attr: str,
    key: str | None,
    convert_fn: Callable[[Any], np.ndarray] | None,
    idx: slice | list,
):
    expected = getattr(dadc[idx], attr)
    if key is not None:
        expected = expected[key]
    if convert_fn is not None:
        expected = convert_fn(expected)

    field = AnnDataField(attr, key, convert_fn)
    actual = field(dadc)[idx]

    np.testing.assert_array_equal(expected, actual)



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

import torch
import torch.distributed as dist


class GatherLayer(torch.autograd.Function):
    """Gather tensors from all process, supporting backward propagation."""

    @staticmethod
    def forward(ctx, input: torch.Tensor) -> tuple[torch.Tensor, ...]:  # type: ignore
        ctx.save_for_backward(input)
        output = [torch.zeros_like(input) for _ in range(dist.get_world_size())]
        dist.all_gather(output, input)
        return tuple(output)

    @staticmethod
    def backward(ctx, *grads) -> torch.Tensor:
        (input,) = ctx.saved_tensors
        grad_out = torch.zeros_like(input)
        grad_out[:] = grads[dist.get_rank()]
        return grad_out



# Copyright Contributors to the Cellarium project.
# SPDX-License-Identifier: BSD-3-Clause

from cellarium.ml.transforms.divide_by_scale import DivideByScale
from cellarium.ml.transforms.filter import Filter
from cellarium.ml.transforms.log1p import Log1p
from cellarium.ml.transforms.normalize_total import NormalizeTotal
from cellarium.ml.transforms.z_score import ZScore

__all__ = ["DivideByScale", "Filter", "Log1p", "NormalizeTotal", "ZScore"]
